import xml.dom.minidom
import xml.parsers.expat

from . import errors
from . import dom


def ParseOsm(data):
    """
    Parse osm data.

    Returns list of dict:

        #!python
        {
            type: node|way|relation,
            data: {}
        }
    """
    try:
        data = xml.dom.minidom.parseString(data)
        data = data.getElementsByTagName("osm")[0]
    except (xml.parsers.expat.ExpatError, IndexError) as e:
        raise errors.XmlResponseInvalidError(
            "The XML response from the OSM API is invalid: %r" % e
        )

    result = []
    for elem in data.childNodes:
        if elem.nodeName == "node":
            result.append({
                "type": elem.nodeName,
                "data": dom.DomParseNode(elem)
            })
        elif elem.nodeName == "way":
            result.append({
                "type": elem.nodeName,
                "data": dom.DomParseWay(elem)
            })
        elif elem.nodeName == "relation":
            result.append({
                "type": elem.nodeName,
                "data": dom.DomParseRelation(elem)
            })
    return result


def ParseOsc(data):
    """
    Parse osc data.

    Returns list of dict:

        #!python
        {
            type: node|way|relation,
            action: create|delete|modify,
            data: {}
        }
    """
    try:
        data = xml.dom.minidom.parseString(data)
        data = data.getElementsByTagName("osmChange")[0]
    except (xml.parsers.expat.ExpatError, IndexError) as e:
        raise errors.XmlResponseInvalidError(
            "The XML response from the OSM API is invalid: %r" % e
        )

    result = []
    for action in data.childNodes:
        if action.nodeName == "#text":
            continue
        for elem in action.childNodes:
            if elem.nodeName == "node":
                result.append({
                    "action": action.nodeName,
                    "type": elem.nodeName,
                    "data": dom.DomParseNode(elem)
                })
            elif elem.nodeName == "way":
                result.append({
                    "action": action.nodeName,
                    "type": elem.nodeName,
                    "data": dom.DomParseWay(elem)
                })
            elif elem.nodeName == "relation":
                result.append({
                    "action": action.nodeName,
                    "type": elem.nodeName,
                    "data": dom.DomParseRelation(elem)
                })
    return result


def ParseNotes(data):
    """
    Parse notes data.

    Returns a list of dict:

        #!python
        [
            {
                'id': integer,
                'action': opened|commented|closed,
                'status': open|closed
                'date_created': creation date
                'date_closed': closing data|None
                'uid': User ID|None
                'user': User name|None
                'comments': {}
            },
            { ... }
        ]
    """
    noteElements = dom.OsmResponseToDom(data, tag="note", allow_empty=True)
    result = []
    for noteElement in noteElements:
        note = dom.DomParseNote(noteElement)
        result.append(note)
    return result



import time
import logging
import requests

from . import errors


logger = logging.getLogger(__name__)


class OsmApiSession:

    MAX_RETRY_LIMIT = 5
    """Maximum retries if a call to the remote API fails (default: 5)"""

    def __init__(self, base_url, created_by, auth=None, session=None):
        self._api = base_url
        self._created_by = created_by
        self._auth = auth

        self._http_session = session
        self._session = self._get_http_session()

    def close(self):
        if self._session:
            self._session.close()

    def _http_request(self, method, path, auth, send, return_value=True):  # noqa
        """
        Returns the response generated by an HTTP request.

        `method` is a HTTP method to be executed
        with the request data. For example: 'GET' or 'POST'.
        `path` is the path to the requested resource relative to the
        base API address stored in self._api. Should start with a
        slash character to separate the URL.
        `auth` is a boolean indicating whether authentication should
        be preformed on this request.
        `send` contains additional data that might be sent in a
        request.
        `return_value` indicates wheter this request should return
        any data or not.

        If the username or password is missing,
        `OsmApi.UsernamePasswordMissingError` is raised.

        If the requested element has been deleted,
        `OsmApi.ElementDeletedApiError` is raised.

        If the requested element can not be found,
        `OsmApi.ElementNotFoundApiError` is raised.

        If the response status code indicates an error,
        `OsmApi.ApiError` is raised.
        """
        msg = (
            "%s %s %s"
            % (time.strftime("%Y-%m-%d %H:%M:%S"), method, path)
        )
        logger.debug(msg)

        # Add API base URL to path
        path = self._api + path

        if auth and not self._auth:
            raise errors.UsernamePasswordMissingError("Username/Password missing")

        response = self._session.request(
            method,
            path,
            data=send
        )
        if response.status_code != 200:
            payload = response.content.strip()
            if response.status_code == 404:
                raise errors.ElementNotFoundApiError(
                    response.status_code,
                    response.reason,
                    payload
                )
            elif response.status_code == 410:
                raise errors.ElementDeletedApiError(
                    response.status_code,
                    response.reason,
                    payload
                )
            raise errors.ApiError(response.status_code, response.reason, payload)
        if return_value and not response.content:
            raise errors.ResponseEmptyApiError(
                response.status_code,
                response.reason,
                ''
            )

        msg = (
            "%s %s %s"
            % (time.strftime("%Y-%m-%d %H:%M:%S"), method, path)
        )
        logger.debug(msg)
        return response.content

    def _http(self, cmd, path, auth, send, return_value=True):  # noqa
        i = 0
        while True:
            i += 1
            try:
                return self._http_request(
                    cmd,
                    path,
                    auth,
                    send,
                    return_value=return_value
                )
            except errors.ApiError as e:
                if e.status >= 500:
                    if i == self.MAX_RETRY_LIMIT:
                        raise
                    if i != 1:
                        self._sleep()
                    self._session = self._get_http_session()
                else:
                    raise
            except Exception as e:
                logger.error(e)
                if i == self.MAX_RETRY_LIMIT:
                    if isinstance(e, errors.OsmApiError):
                        raise
                    raise errors.MaximumRetryLimitReachedError(
                        "Give up after %s retries" % i
                    )
                if i != 1:
                    self._sleep()
                self._session = self._get_http_session()

    def _get_http_session(self):
        """
        Creates a requests session for connection pooling.
        """
        if self._http_session:
            session = self._http_session
        else:
            session = requests.Session()

        session.auth = self._auth
        session.headers.update({
            'user-agent': self._created_by
        })
        return session

    def _sleep(self):
        time.sleep(5)

    def _get(self, path):
        return self._http('GET', path, False, None)

    def _put(self, path, data, return_value=True):
        return self._http('PUT', path, True, data, return_value=return_value)

    def _post(self, path, data, optionalAuth=False, forceAuth=False):
        # the Notes API allows certain POSTs by non-authenticated users
        auth = (optionalAuth and self._auth)
        if forceAuth:
            auth = True
        return self._http('POST', path, auth, data)

    def _delete(self, path, data):
        return self._http('DELETE', path, True, data)



import osmapi
import pytest
import mock
import responses
import os
import re

__location__ = os.path.realpath(
    os.path.join(
        os.getcwd(),
        os.path.dirname(__file__)
    )
)


@pytest.fixture
def file_content():
    def _file_content(filename):
        path = os.path.join(__location__, 'fixtures', filename)
        if not os.path.exists(path):
            return ""
        with open(path) as f:
            return f.read()

    return _file_content


@pytest.fixture
def api():
    api_base = "http://api06.dev.openstreetmap.org"
    api = osmapi.OsmApi(
        api=api_base
    )
    api._session._sleep = mock.Mock()

    yield api
    api.close()


@pytest.fixture
def auth_api():
    api_base = "http://api06.dev.openstreetmap.org"
    api = osmapi.OsmApi(
        api=api_base,
        username='testuser',
        password='testpassword'
    )
    api._session._sleep = mock.Mock()

    yield api
    api.close()


@pytest.fixture
def prod_api():
    api_base = "https://www.openstreetmap.org"
    api = osmapi.OsmApi(
        api=api_base,
        username='testuser',
        password='testpassword'
    )
    api._session._sleep = mock.Mock()

    yield api
    api.close()


@pytest.fixture
def mocked_responses():
    with responses.RequestsMock() as rsps:
        yield rsps


@pytest.fixture
def add_response(mocked_responses, file_content, request):
    def _add_response(method, path=None, filename=None, body=None, status=200):
        if not filename:
            # use testname by default
            filename = f"{request.node.originalname}.xml"

        if path:
            url = f'http://api06.dev.openstreetmap.org/api/0.6{path}'
        else:
            url = re.compile(r'http:\/\/api06\.dev\.openstreetmap\.org.*')

        if not body:
            body = file_content(filename)
        mocked_responses.add(
            method,
            url=url,
            body=body,
            status=status
        )
        return mocked_responses
    return _add_response



from __future__ import (unicode_literals, absolute_import)
from . import osmapi_test
import osmapi
import mock
import datetime


class TestOsmApiRelation(osmapi_test.TestOsmApi):
    def test_RelationGet(self):
        self._session_mock()

        result = self.api.RelationGet(321)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], 'GET')
        self.assertEqual(args[1], self.api_base + '/api/0.6/relation/321')

        self.assertEqual(result, {
            'id': 321,
            'changeset': 434,
            'uid': 12,
            'timestamp': datetime.datetime(2009, 9, 15, 22, 24, 25),
            'visible': True,
            'version': 1,
            'user': 'green525',
            'tag': {
                'admin_level': '9',
                'boundary': 'administrative',
                'type': 'multipolygon',
            },
            'member': [
                {
                    'ref': 6908,
                    'role': 'outer',
                    'type': 'way'
                },
                {
                    'ref': 6352,
                    'role': 'outer',
                    'type': 'way'
                },
                {
                    'ref': 5669,
                    'role': 'outer',
                    'type': 'way'
                },
                {
                    'ref': 5682,
                    'role': 'outer',
                    'type': 'way'
                },
                {
                    'ref': 6909,
                    'role': 'outer',
                    'type': 'way'
                },
                {
                    'ref': 6355,
                    'role': 'outer',
                    'type': 'way'
                },
                {
                    'ref': 6910,
                    'role': 'outer',
                    'type': 'way'
                },
                {
                    'ref': 6911,
                    'role': 'outer',
                    'type': 'way'
                },
                {
                    'ref': 6912,
                    'role': 'outer',
                    'type': 'way'
                }
            ]
        })

    def test_RelationGet_with_version(self):
        self._session_mock()

        result = self.api.RelationGet(765, 2)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], 'GET')
        self.assertEqual(args[1], self.api_base + '/api/0.6/relation/765/2')

        self.assertEqual(result['id'], 765)
        self.assertEqual(result['changeset'], 41378)
        self.assertEqual(result['user'], 'metaodi')
        self.assertEqual(result['tag']['source'], 'test')

    def test_RelationCreate(self):
        self._session_mock(auth=True)

        # setup mock
        self.api.ChangesetCreate = mock.Mock(
            return_value=3333
        )
        self.api._CurrentChangesetId = 3333

        test_relation = {
            'tag': {
                'type': 'test',
            },
            'member': [
                {
                    'ref': 6908,
                    'role': 'outer',
                    'type': 'way'
                },
                {
                    'ref': 6352,
                    'role': 'point',
                    'type': 'node'
                },
            ]
        }

        cs = self.api.ChangesetCreate({
            'comment': 'This is a test relation'
        })
        self.assertEqual(cs, 3333)

        result = self.api.RelationCreate(test_relation)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], 'PUT')
        self.assertEqual(args[1], self.api_base + '/api/0.6/relation/create')

        self.assertEqual(result['id'], 8989)
        self.assertEqual(result['version'], 1)
        self.assertEqual(result['member'], test_relation['member'])
        self.assertEqual(result['tag'], test_relation['tag'])

    def test_RelationCreate_existing_node(self):
        # setup mock
        self.api.ChangesetCreate = mock.Mock(
            return_value=1111
        )
        self.api._CurrentChangesetId = 1111

        test_relation = {
            'id': 456,
            'tag': {
                'type': 'test',
            },
            'member': [
                {
                    'ref': 6908,
                    'role': 'outer',
                    'type': 'way'
                },
                {
                    'ref': 6352,
                    'role': 'point',
                    'type': 'node'
                },
            ]
        }

        with self.assertRaisesRegex(
                osmapi.OsmTypeAlreadyExistsError,
                'This relation already exists'):
            self.api.RelationCreate(test_relation)

    def test_RelationUpdate(self):
        self._session_mock(auth=True)

        # setup mock
        self.api.ChangesetCreate = mock.Mock(
            return_value=3333
        )
        self.api._CurrentChangesetId = 3333

        test_relation = {
            'id': 8989,
            'tag': {
                'type': 'test update',
            },
            'member': [
                {
                    'ref': 6908,
                    'role': 'outer',
                    'type': 'way'
                }
            ]
        }

        cs = self.api.ChangesetCreate({
            'comment': 'This is a test relation'
        })
        self.assertEqual(cs, 3333)

        result = self.api.RelationUpdate(test_relation)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], 'PUT')
        self.assertEqual(args[1], self.api_base + '/api/0.6/relation/8989')

        self.assertEqual(result['id'], 8989)
        self.assertEqual(result['version'], 42)
        self.assertEqual(result['member'], test_relation['member'])
        self.assertEqual(result['tag'], test_relation['tag'])

    def test_RelationDelete(self):
        self._session_mock(auth=True)

        # setup mock
        self.api.ChangesetCreate = mock.Mock(
            return_value=3333
        )
        self.api._CurrentChangesetId = 3333

        test_relation = {
            'id': 8989
        }

        cs = self.api.ChangesetCreate({
            'comment': 'This is a test relation'
        })
        self.assertEqual(cs, 3333)

        result = self.api.RelationDelete(test_relation)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], 'DELETE')
        self.assertEqual(args[1], self.api_base + '/api/0.6/relation/8989')

        self.assertEqual(result['id'], 8989)
        self.assertEqual(result['version'], 43)

    def test_RelationHistory(self):
        self._session_mock()

        result = self.api.RelationHistory(2470397)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], 'GET')
        self.assertEqual(
            args[1],
            f'{self.api_base}/api/0.6/relation/2470397/history'
        )

        self.assertEqual(len(result), 2)
        self.assertEqual(result[1]['id'], 2470397)
        self.assertEqual(result[1]['version'], 1)
        self.assertEqual(
            result[1]['tag'], {
                'restriction': 'only_straight_on',
                'type': 'restriction',
            }
        )
        self.assertEqual(result[2]['version'], 2)

    def test_RelationRelations(self):
        self._session_mock()

        result = self.api.RelationRelations(1532552)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], 'GET')
        self.assertEqual(
            args[1],
            f'{self.api_base}/api/0.6/relation/1532552/relations'
        )

        self.assertEqual(len(result), 1)
        self.assertEqual(result[0]['id'], 1532553)
        self.assertEqual(result[0]['version'], 85)
        self.assertEqual(len(result[0]['member']), 120)
        self.assertEqual(result[0]['tag']['type'], 'network')
        self.assertEqual(
            result[0]['tag']['name'],
            'Aargauischer Radroutennetz'
        )

    def test_RelationRelationsUnusedElement(self):
        self._session_mock()

        result = self.api.RelationRelations(1532552)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], 'GET')
        self.assertEqual(
            args[1],
            f'{self.api_base}/api/0.6/relation/1532552/relations'
        )

        self.assertEqual(len(result), 0)
        self.assertIsInstance(result, list)

    def test_RelationFull(self):
        self._session_mock()

        result = self.api.RelationFull(2470397)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], 'GET')
        self.assertEqual(
            args[1],
            f'{self.api_base}/api/0.6/relation/2470397/full'
        )

        self.assertEqual(len(result), 11)
        self.assertEqual(result[1]['data']['id'], 101142277)
        self.assertEqual(result[1]['data']['version'], 8)
        self.assertEqual(result[1]['type'], 'node')
        self.assertEqual(result[10]['data']['id'], 2470397)
        self.assertEqual(result[10]['data']['version'], 2)
        self.assertEqual(result[10]['type'], 'relation')

    def test_RelationsGet(self):
        self._session_mock()

        result = self.api.RelationsGet([1532552, 1532553])

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], 'GET')
        self.assertEqual(
            args[1],
            f'{self.api_base}/api/0.6/relations?relations=1532552,1532553'
        )

        self.assertEqual(len(result), 2)
        self.assertEqual(result[1532553]['id'], 1532553)
        self.assertEqual(result[1532553]['version'], 85)
        self.assertEqual(result[1532553]['user'], 'SimonPoole')
        self.assertEqual(result[1532552]['id'], 1532552)
        self.assertEqual(result[1532552]['visible'], True)
        self.assertEqual(result[1532552]['tag']['route'], 'bicycle')

    def test_RelationFull_with_deleted_relation(self):
        self._session_mock(filenames=[], status=410)

        with self.assertRaises(osmapi.ElementDeletedApiError) as context:
            self.api.RelationFull(2911456)
        self.assertEqual(410, context.exception.status)



def _XmlBuild(ElementType, ElementData, WithHeaders=True, data=None):  # noqa
    xml = ""
    if WithHeaders:
        xml += "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
        xml += "<osm version=\"0.6\" generator=\""
        xml += data._created_by + "\">\n"

    # <element attr="val">
    xml += "  <" + ElementType
    if "id" in ElementData:
        xml += " id=\"" + str(ElementData["id"]) + "\""
    if "lat" in ElementData:
        xml += " lat=\"" + str(ElementData["lat"]) + "\""
    if "lon" in ElementData:
        xml += " lon=\"" + str(ElementData["lon"]) + "\""
    if "version" in ElementData:
        xml += " version=\"" + str(ElementData["version"]) + "\""
    visible_str = str(ElementData.get("visible", True)).lower()
    xml += " visible=\"" + visible_str + "\""
    if ElementType in ["node", "way", "relation"]:
        xml += " changeset=\"" + str(data._CurrentChangesetId) + "\""
    xml += ">\n"

    # <tag... />
    for k, v in ElementData.get("tag", {}).items():
        xml += "    <tag k=\"" + _XmlEncode(k)
        xml += "\" v=\"" + _XmlEncode(v) + "\"/>\n"

    # <member... />
    for member in ElementData.get("member", []):
        xml += "    <member type=\"" + member["type"]
        xml += "\" ref=\"" + str(member["ref"])
        xml += "\" role=\"" + _XmlEncode(member["role"])
        xml += "\"/>\n"

    # <nd... />
    for ref in ElementData.get("nd", []):
        xml += "    <nd ref=\"" + str(ref) + "\"/>\n"

    # </element>
    xml += "  </" + ElementType + ">\n"

    if WithHeaders:
        xml += "</osm>\n"

    return xml.encode("utf8")


def _XmlEncode(text):
    return (
        text
        .replace("&", "&amp;")
        .replace("\"", "&quot;")
        .replace("<", "&lt;")
        .replace(">", "&gt;")
    )


def _GetXmlValue(DomElement, tag):
    try:
        elem = DomElement.getElementsByTagName(tag)[0]
        return elem.firstChild.nodeValue
    except Exception:
        return None



import osmapi
from dotenv import load_dotenv, find_dotenv
import os
from pprint import pprint

load_dotenv(find_dotenv())
user = os.getenv('OSM_USER')
pw = os.getenv('OSM_PASS')

api = osmapi.OsmApi(api="https://api06.dev.openstreetmap.org", username=user, password=pw)
empty_notes = api.NotesGet(-93.8472901, 35.9763601, -80, 36.176360100000004, limit=1, closed=0)
pprint(empty_notes)


# create note and then search for it
note = api.NoteCreate({
    'lat':  47.3383501,
    'lon': 8.5339522,
    'text': 'test note',
})
test_notes = api.NotesGet(8.527504, 47.337063, 8.540679, 47.341673, limit=1, closed=0)
pprint(test_notes)


api.NoteComment(note['id'], "Another comment")
api.NoteClose(note['id'], "Close this test note")


# try to close an already closed note
try:
    api.NoteClose(note['id'], "Close the note again")
except osmapi.NoteAlreadyClosedApiError:
    print('')
    print(f"The note {note['id']} has already been closed")

# try to comment on closed note
try:
    api.NoteComment(note['id'], "Just a comment")
except osmapi.NoteAlreadyClosedApiError:
    print('')
    print(f"The note {note['id']} is closed, comment no longer possible")



import osmapi
from dotenv import load_dotenv, find_dotenv
import os

load_dotenv(find_dotenv())
user = os.getenv('OSM_USER')
pw = os.getenv('OSM_PASS')

api = osmapi.OsmApi(api="https://api06.dev.openstreetmap.org", username=user, password=pw)
with api.Changeset({u"comment": u"My first test"}) as changeset_id:
    print(f"Part of Changeset {changeset_id}")
    node1 = api.NodeCreate({u"lon": 1, u"lat": 1, u"tag": {}})
    print(node1)
    node2 = api.NodeCreate({u"lon": 2, u"lat": 2, u"tag": {}})
    print(node2)
    way = api.WayCreate({
        'nd': [
            node1['id'],
            node2['id'],
        ],
        'tag': {
            'highway': 'unclassified',
            'name': 'Osmapi Street',
        }
    })
    print(way)



from __future__ import unicode_literals
from osmapi import OsmApi
import mock
import os
import unittest

__location__ = os.path.realpath(
    os.path.join(
        os.getcwd(),
        os.path.dirname(__file__)
    )
)


class TestOsmApi(unittest.TestCase):
    def setUp(self):
        self.api_base = "http://api06.dev.openstreetmap.org"
        self.api = OsmApi(
            api=self.api_base
        )
        self.maxDiff = None
        print(self._testMethodName)
        print(self.api)

    def _session_mock(self, auth=False, filenames=None, status=200):
        response_mock = mock.Mock()
        response_mock.status_code = status
        return_values = self._return_values(filenames)
        print(filenames)
        print(return_values)
        assert len(return_values) < 2
        if return_values:
            response_mock.content = return_values[0]

        self.session_mock = mock.Mock()
        self.session_mock.request = mock.Mock(return_value=response_mock)

        if auth:
            self.api = OsmApi(
                api=self.api_base,
                username='testuser',
                password='testpassword',
                session=self.session_mock
            )
        else:
            self.api = OsmApi(
                api=self.api_base,
                session=self.session_mock
            )

        self.api._get_http_session = mock.Mock(return_value=self.session_mock)
        self.api._session._sleep = mock.Mock()

    def _return_values(self, filenames):
        if filenames is None:
            filenames = [self._testMethodName + ".xml"]

        return_values = []
        for filename in filenames:
            path = os.path.join(
                __location__,
                'fixtures',
                filename
            )
            try:
                with open(path) as file:
                    return_values.append(file.read())
            except Exception:
                pass
        return return_values

    def teardown(self):
        pass

    def test_constructor(self):
        self.assertTrue(isinstance(self.api, OsmApi))



# -*- coding: utf-8 -*-

class OsmApiError(Exception):
    """
    General OsmApi error class to provide a superclass for all other errors
    """


class MaximumRetryLimitReachedError(OsmApiError):
    """
    Error when the maximum amount of retries is reached and we have to give up
    """


class UsernamePasswordMissingError(OsmApiError):
    """
    Error when username or password is missing for an authenticated request
    """
    pass


class NoChangesetOpenError(OsmApiError):
    """
    Error when an operation requires an open changeset, but currently
    no changeset _is_ open
    """
    pass


class ChangesetAlreadyOpenError(OsmApiError):
    """
    Error when a user tries to open a changeset when there is already
    an open changeset
    """
    pass


class OsmTypeAlreadyExistsError(OsmApiError):
    """
    Error when a user tries to create an object that already exsits
    """
    pass


class XmlResponseInvalidError(OsmApiError):
    """
    Error if the XML response from the OpenStreetMap API is invalid
    """


class ApiError(OsmApiError):
    """
    Error class, is thrown when an API request fails
    """

    def __init__(self, status, reason, payload):
        self.status = status
        """HTTP error code"""

        self.reason = reason
        """Error message"""

        self.payload = payload
        """Payload of API when this error occured"""

    def __str__(self):
        return (
            "Request failed: %s - %s - %s"
            % (str(self.status), self.reason, self.payload)
        )


class AlreadySubscribedApiError(ApiError):
    """
    Error when a user tries to subscribe to a changeset
    that she is already subscribed to
    """
    pass


class NotSubscribedApiError(ApiError):
    """
    Error when user tries to unsubscribe from a changeset
    that he is not subscribed to
    """
    pass


class ElementDeletedApiError(ApiError):
    """
    Error when the requested element is deleted
    """
    pass


class ElementNotFoundApiError(ApiError):
    """
    Error if the the requested element was not found
    """


class ResponseEmptyApiError(ApiError):
    """
    Error when the response to the request is empty
    """
    pass


class ChangesetClosedApiError(ApiError):
    """
    Error if the the changeset in question has already been closed
    """


class NoteAlreadyClosedApiError(ApiError):
    """
    Error if the the note in question has already been closed
    """


class VersionMismatchApiError(ApiError):
    """
    Error if the provided version does not match the database version
    of the element
    """


class PreconditionFailedApiError(ApiError):
    """
    Error if the precondition of the operation was not met:
    - When a way has nodes that do not exist or are not visible
    - When a relation has elements that do not exist or are not visible
    - When a node/way/relation is still used in a way/relation
    """



import osmapi
import xmltodict
import datetime
import pytest
from responses import GET, PUT, POST


def xmltosorteddict(xml):
    xml_dict = xmltodict.parse(xml, dict_constructor=dict)
    return xml_dict


def test_Changeset_contextmanager(auth_api, add_response):
    # Setup mock
    resp = add_response(PUT, '/changeset/create', filename='test_Changeset_create.xml')
    resp = add_response(PUT, '/node/create', filename='test_Changeset_create_node.xml')
    resp = add_response(PUT, '/changeset/1414/close', filename='test_Changeset_close.xml')

    test_node = {
        'lat': 47.123,
        'lon': 8.555,
        'tag': {
            'amenity': 'place_of_worship',
            'religion': 'pastafarian'
        }
    }

    # use context manager
    with auth_api.Changeset() as changeset_id:
        assert changeset_id == 1414

        # add test node
        node = auth_api.NodeCreate(test_node)
        assert node['id'] == 7272

    # check requests
    assert len(resp.calls) == 3


def test_ChangesetGet(api, add_response):
    # Setup mock
    add_response(GET, '/changeset/123')

    # Call
    result = api.ChangesetGet(123)

    test_changeset = {
        'id': 123,
        'closed_at': datetime.datetime(2009, 9, 7, 22, 57, 37),
        'created_at': datetime.datetime(2009, 9, 7, 21, 57, 36),
        'discussion': [],
        'max_lat': '52.4710193',
        'max_lon': '-1.4831815',
        'min_lat': '45.9667901',
        'min_lon': '-1.4998534',
        'open': False,
        'user': 'randomjunk',
        'uid': 3,
        'tag': {
            'comment': 'correct node bug',
            'created_by': 'Potlatch 1.2a',
        },
    }
    assert result == test_changeset


def test_ChangesetUpdate(auth_api, add_response):
    # Setup mock
    resp = add_response(PUT, '/changeset/create', filename='test_ChangesetCreate.xml')
    resp = add_response(PUT, '/changeset/4321', filename='test_ChangesetUpdate.xml')

    # Call
    result = auth_api.ChangesetCreate()
    assert result == 4321

    result = auth_api.ChangesetUpdate({'test': 'foobar'})
    changeset_xml = xmltosorteddict(
        b'<?xml version="1.0" encoding="UTF-8"?>\n'
        b'<osm version="0.6" generator="osmapi/3.1.0">\n'
        b'  <changeset visible="true">\n'
        b'    <tag k="test" v="foobar"/>\n'
        b'    <tag k="created_by" v="osmapi/3.1.0"/>\n'
        b'  </changeset>\n'
        b'</osm>\n'
    )
    assert xmltosorteddict(resp.calls[1].request.body) == changeset_xml
    assert result == 4321


def test_ChangesetUpdate_with_created_by(auth_api, add_response):
    # Setup mock
    resp = add_response(PUT, '/changeset/create', filename='test_ChangesetCreate.xml')
    resp = add_response(PUT, '/changeset/4321', filename='test_ChangesetUpdate.xml')

    # Call
    result = auth_api.ChangesetCreate()
    assert result == 4321

    result = auth_api.ChangesetUpdate(
        {
            'test': 'foobar',
            'created_by': 'MyTestOSMApp'
        }
    )
    changeset_xml = xmltosorteddict(
        b'<?xml version="1.0" encoding="UTF-8"?>\n'
        b'<osm version="0.6" generator="osmapi/3.1.0">\n'
        b'  <changeset visible="true">\n'
        b'    <tag k="test" v="foobar"/>\n'
        b'    <tag k="created_by" v="MyTestOSMApp"/>\n'
        b'  </changeset>\n'
        b'</osm>\n'
    )
    assert xmltosorteddict(resp.calls[1].request.body) == changeset_xml
    assert result == 4321


def test_ChangesetUpdate_wo_changeset(auth_api):
    with pytest.raises(osmapi.NoChangesetOpenError) as execinfo:
        auth_api.ChangesetUpdate({'test': 'foobar'})
    assert str(execinfo.value) == 'No changeset currently opened'


def test_ChangesetCreate(auth_api, add_response):
    resp = add_response(PUT, '/changeset/create')
    result = auth_api.ChangesetCreate(
        {
            'foobar': 'A new test changeset'
        }
    )
    assert result == 4321

    changeset_xml = xmltosorteddict(
        b'<?xml version="1.0" encoding="UTF-8"?>\n'
        b'<osm version="0.6" generator="osmapi/3.1.0">\n'
        b'  <changeset visible="true">\n'
        b'    <tag k="foobar" v="A new test changeset"/>\n'
        b'    <tag k="created_by" v="osmapi/3.1.0"/>\n'
        b'  </changeset>\n'
        b'</osm>\n'
    )
    assert xmltosorteddict(resp.calls[0].request.body) == changeset_xml


def test_ChangesetCreate_with_created_by(auth_api, add_response):
    resp = add_response(PUT, '/changeset/create')

    result = auth_api.ChangesetCreate(
        {
            'foobar': 'A new test changeset',
            'created_by': 'CoolTestApp',
        }
    )
    assert result == 1234

    changeset_xml = xmltosorteddict(
        b'<?xml version="1.0" encoding="UTF-8"?>\n'
        b'<osm version="0.6" generator="osmapi/3.1.0">\n'
        b'  <changeset visible="true">\n'
        b'    <tag k="foobar" v="A new test changeset"/>\n'
        b'    <tag k="created_by" v="CoolTestApp"/>\n'
        b'  </changeset>\n'
        b'</osm>\n'
    )
    assert xmltosorteddict(resp.calls[0].request.body) == changeset_xml


def test_ChangesetCreate_with_open_changeset(auth_api, add_response):
    add_response(PUT, '/changeset/create')

    auth_api.ChangesetCreate(
        {
            'test': 'an already open changeset',
        }
    )

    with pytest.raises(osmapi.ChangesetAlreadyOpenError) as execinfo:
        auth_api.ChangesetCreate({'test': 'foobar'})
    assert str(execinfo.value) == 'Changeset already opened'


def test_ChangesetCreate_with_prod_api_and_test_comment(prod_api):
    with pytest.raises(osmapi.OsmApiError) as execinfo:
        prod_api.ChangesetCreate(
            {
                'comment': 'My first test',
            }
        )
    assert str(execinfo.value) == 'DO NOT CREATE test changesets on the production server'


def test_ChangesetClose(auth_api, add_response):
    # setup mock
    resp = add_response(PUT, '/changeset/create', filename='test_Changeset_create.xml')
    resp = add_response(PUT, '/changeset/1414/close')

    # Call
    auth_api.ChangesetCreate()
    auth_api.ChangesetClose()

    assert '/api/0.6/changeset/1414/close' in resp.calls[1].request.url


def test_ChangesetClose_with_no_changeset(auth_api):
    with pytest.raises(osmapi.NoChangesetOpenError) as execinfo:
        auth_api.ChangesetClose()
    assert str(execinfo.value) == 'No changeset currently opened'


def test_ChangesetUpload_create_node(auth_api, add_response):
    # Setup
    resp = add_response(PUT, '/changeset/create', body='4444')
    resp = add_response(POST, '/changeset/4444/upload')

    changesdata = [
        {
            'type': 'node',
            'action': 'create',
            'data': {
                'lat': 47.123,
                'lon': 8.555,
                'tag': {
                    'amenity': 'place_of_worship',
                    'religion': 'pastafarian'
                }
            }
        }
    ]

    upload_xml = xmltosorteddict(
        b'<?xml version="1.0" encoding="UTF-8"?>\n'
        b'<osmChange version="0.6" generator="osmapi/3.1.0">\n'
        b'<create>\n'
        b'  <node lat="47.123" lon="8.555" visible="true" '
        b'changeset="4444">\n'
        b'    <tag k="amenity" v="place_of_worship"/>\n'
        b'    <tag k="religion" v="pastafarian"/>\n'
        b'  </node>\n'
        b'</create>\n'
        b'</osmChange>'
    )

    # Call
    auth_api.ChangesetCreate()
    result = auth_api.ChangesetUpload(changesdata)

    # Assert
    assert xmltosorteddict(resp.calls[1].request.body) == upload_xml
    assert result[0]['type'] == changesdata[0]['type']
    assert result[0]['action'] == changesdata[0]['action']

    data = result[0]['data']
    assert data['lat'] == changesdata[0]['data']['lat']
    assert data['lon'] == changesdata[0]['data']['lon']
    assert data['tag'] == changesdata[0]['data']['tag']
    assert data['id'] == 4295832900
    assert result[0]['data']['version'] == 1


def test_ChangesetUpload_modify_way(auth_api, add_response):
    # setup mock
    resp = add_response(PUT, '/changeset/create', body='4444')
    resp = add_response(POST, '/changeset/4444/upload')

    changesdata = [
        {
            'type': 'way',
            'action': 'modify',
            'data': {
                'id': 4294967296,
                'version': 2,
                'nd': [
                    4295832773,
                    4295832773,
                    4294967304,
                    4294967303,
                    4294967300,
                    4608751,
                    4294967305,
                    4294967302,
                    8548430,
                    4294967296,
                    4294967301,
                    4294967298,
                    4294967306,
                    7855737,
                    4294967297,
                    4294967299
                ],
                'tag': {
                    'highway': 'secondary',
                    'name': 'Stansted Road'
                }
            }
        }
    ]

    upload_xml = xmltosorteddict(
        b'<?xml version="1.0" encoding="UTF-8"?>\n'
        b'<osmChange version="0.6" generator="osmapi/3.1.0">\n'
        b'<modify>\n'
        b'  <way id="4294967296" version="2" visible="true" '
        b'changeset="4444">\n'
        b'    <tag k="highway" v="secondary"/>\n'
        b'    <tag k="name" v="Stansted Road"/>\n'
        b'    <nd ref="4295832773"/>\n'
        b'    <nd ref="4295832773"/>\n'
        b'    <nd ref="4294967304"/>\n'
        b'    <nd ref="4294967303"/>\n'
        b'    <nd ref="4294967300"/>\n'
        b'    <nd ref="4608751"/>\n'
        b'    <nd ref="4294967305"/>\n'
        b'    <nd ref="4294967302"/>\n'
        b'    <nd ref="8548430"/>\n'
        b'    <nd ref="4294967296"/>\n'
        b'    <nd ref="4294967301"/>\n'
        b'    <nd ref="4294967298"/>\n'
        b'    <nd ref="4294967306"/>\n'
        b'    <nd ref="7855737"/>\n'
        b'    <nd ref="4294967297"/>\n'
        b'    <nd ref="4294967299"/>\n'
        b'  </way>\n'
        b'</modify>\n'
        b'</osmChange>'
    )

    # Call
    auth_api.ChangesetCreate()
    result = auth_api.ChangesetUpload(changesdata)

    # Assert
    assert xmltosorteddict(resp.calls[1].request.body) == upload_xml

    assert result[0]['type'] == changesdata[0]['type']
    assert result[0]['action'] == changesdata[0]['action']

    data = result[0]['data']
    assert data['nd'] == changesdata[0]['data']['nd']
    assert data['tag'] == changesdata[0]['data']['tag']
    assert data['id'] == 4294967296
    assert data['version'] == 3


def test_ChangesetUpload_delete_relation(auth_api, add_response):
    # setup mock
    resp = add_response(PUT, '/changeset/create', body='4444')
    resp = add_response(POST, '/changeset/4444/upload')

    changesdata = [
        {
            'type': 'relation',
            'action': 'delete',
            'data': {
                'id': 676,
                'version': 2,
                'member': [
                    {
                        'ref': 4799,
                        'role': 'outer',
                        'type': 'way'
                    },
                    {
                        'ref': 9391,
                        'role': 'outer',
                        'type': 'way'
                    },
                ],
                'tag': {
                    'admin_level': '9',
                    'boundary': 'administrative',
                    'type': 'multipolygon'
                }
            }
        }
    ]

    upload_xml = xmltosorteddict(
        b'<?xml version="1.0" encoding="UTF-8"?>\n'
        b'<osmChange version="0.6" generator="osmapi/3.1.0">\n'
        b'<delete>\n'
        b'  <relation id="676" version="2" visible="true" '
        b'changeset="4444">\n'
        b'    <tag k="admin_level" v="9"/>\n'
        b'    <tag k="boundary" v="administrative"/>\n'
        b'    <tag k="type" v="multipolygon"/>\n'
        b'    <member type="way" ref="4799" role="outer"/>\n'
        b'    <member type="way" ref="9391" role="outer"/>\n'
        b'  </relation>\n'
        b'</delete>\n'
        b'</osmChange>'
    )

    # Call
    auth_api.ChangesetCreate()
    result = auth_api.ChangesetUpload(changesdata)

    # Assert
    assert xmltosorteddict(resp.calls[1].request.body) == upload_xml
    assert result[0]['type'] == changesdata[0]['type']
    assert result[0]['action'] == changesdata[0]['action']

    data = result[0]['data']
    assert data['member'], changesdata[0]['data']['member']
    assert data['tag'] == changesdata[0]['data']['tag']
    assert data['id'] == 676
    assert 'version' not in data


def test_ChangesetUpload_invalid_response(auth_api, add_response):
    # setup mock
    add_response(PUT, '/changeset/create', body='4444')
    add_response(POST, '/changeset/4444/upload', body='4444')

    changesdata = [
        {
            'type': 'relation',
            'action': 'delete',
            'data': {
                'id': 676,
                'version': 2,
                'member': [
                    {
                        'ref': 4799,
                        'role': 'outer',
                        'type': 'way'
                    },
                    {
                        'ref': 9391,
                        'role': 'outer',
                        'type': 'way'
                    },
                ],
                'tag': {
                    'admin_level': '9',
                    'boundary': 'administrative',
                    'type': 'multipolygon'
                }
            }
        }
    ]

    # Call + assert
    auth_api.ChangesetCreate()
    with pytest.raises(osmapi.XmlResponseInvalidError) as execinfo:
        auth_api.ChangesetUpload(changesdata)
    assert 'The XML response from the OSM API is invalid' in str(execinfo.value)


def test_ChangesetUpload_no_auth(api):
    changesdata = [
        {
            'type': 'node',
            'action': 'create',
            'data': {
                'lat': 47.123,
                'lon': 8.555,
                'tag': {
                    'amenity': 'place_of_worship',
                    'religion': 'pastafarian'
                }
            }
        }
    ]

    with pytest.raises(osmapi.UsernamePasswordMissingError) as execinfo:
        api.ChangesetUpload(changesdata)
    assert str(execinfo.value) == "Username/Password missing"


def test_ChangesetDownload(api, add_response):
    # Setup mock
    add_response(GET, '/changeset/23123/download')

    # Call
    result = api.ChangesetDownload(23123)

    # Assertion
    assert len(result) == 16
    assert result[1] == (
        {
            'action': 'create',
            'type': 'node',
            'data': {
                'changeset': 23123,
                'id': 4295668171,
                'lat': 46.4909781,
                'lon': 11.2743295,
                'tag': {
                    'highway': 'traffic_signals'
                },
                'timestamp': datetime.datetime(2013, 5, 14, 10, 33, 4),
                'uid': 1178,
                'user': 'tyrTester06',
                'version': 1,
                'visible': True
            }
        }
    )


def test_ChangesetDownload_invalid_response(api, add_response):
    add_response(GET, '/changeset/23123/download')
    with pytest.raises(osmapi.XmlResponseInvalidError) as execinfo:
        api.ChangesetDownload(23123)
    assert 'The XML response from the OSM API is invalid' in str(execinfo.value)


def test_ChangesetDownloadContainingUnicode(api, add_response):
    add_response(GET, '/changeset/37393499/download')

    # This changeset contains unicode tag values
    # Note that the fixture data has been reduced from the
    # original from openstreetmap.org
    result = api.ChangesetDownload(37393499)

    assert len(result) == 2
    assert result[1] == (
        {
            'action': 'create',
            'type': 'way',
            'data': {
                'changeset': 37393499,
                'id': 399491497,
                'nd': [4022271571, 4022271567, 4022271565],
                'tag': {'highway': 'service',
                        # UTF-8 encoded 'LATIN SMALL LETTER O WITH STROKE'
                        # Aka. 0xf8 in latin-1/ISO 8859-1
                        'name': b'S\xc3\xb8nderskovvej'.decode('utf-8'),
                        'service': 'driveway'},
                'timestamp': datetime.datetime(2016, 2, 23, 16, 55, 35),
                'uid': 328556,
                'user': 'InternationalUser',
                'version': 1,
                'visible': True
            }
        }
    )


def test_ChangesetsGet(api, add_response):
    resp = add_response(GET, '/changesets')

    result = api.ChangesetsGet(
        only_closed=True,
        username='metaodi'
    )

    assert resp.calls[0].request.params == {'display_name': 'metaodi', 'closed': '1'}
    assert len(result) == 10
    assert result[41417] == ({
        'closed_at': datetime.datetime(2014, 4, 29, 20, 25, 1),
        'created_at': datetime.datetime(2014, 4, 29, 20, 25, 1),
        'id': 41417,
        'discussion': [],
        'max_lat': '58.8997467',
        'max_lon': '22.7364427',
        'min_lat': '58.8501594',
        'min_lon': '22.6984333',
        'open': False,
        'tag': {
            'comment': 'Test delete of relation',
            'created_by': 'iD 1.3.9',
            'imagery_used': 'Bing'
        },
        'uid': 1841,
        'user': 'metaodi'
    })


def test_ChangesetGetWithComment(api, add_response):
    resp = add_response(GET, '/changeset/52924')

    result = api.ChangesetGet(52924, include_discussion=True)

    assert resp.calls[0].request.params == {'include_discussion': 'true'}
    assert result == {
        'id': 52924,
        'closed_at': datetime.datetime(2015, 1, 1, 14, 54, 2),
        'created_at': datetime.datetime(2015, 1, 1, 14, 54, 1),
        'comments_count': 3,
        'max_lat': '58.3369242',
        'max_lon': '25.8829107',
        'min_lat': '58.336813',
        'min_lon': '25.8823273',
        'discussion': [
            {
                'date':  datetime.datetime(2015, 1, 1, 18, 56, 48),
                'text': 'test',
                'uid': 1841,
                'user': 'metaodi',
            },
            {
                'date':  datetime.datetime(2015, 1, 1, 18, 58, 3),
                'text': 'another comment',
                'uid': 1841,
                'user': 'metaodi',
            },
            {
                'date':  datetime.datetime(2015, 1, 1, 19, 16, 5),
                'text': 'hello',
                'uid': 1841,
                'user': 'metaodi',
            },
        ],
        'open': False,
        'user': 'metaodi',
        'uid': 1841,
        'tag': {
            'comment': 'My test',
            'created_by': 'osmapi/0.4.1',
        },
    }


def test_ChangesetComment(auth_api, add_response):
    resp = add_response(POST, '/changeset/123/comment')

    result = auth_api.ChangesetComment(
        123,
        comment="test comment"
    )

    assert resp.calls[0].request.body == "text=test+comment"
    assert result == {
        'id': 123,
        'closed_at': datetime.datetime(2009, 9, 7, 22, 57, 37),
        'created_at': datetime.datetime(2009, 9, 7, 21, 57, 36),
        'discussion': [],
        'max_lat': '52.4710193',
        'max_lon': '-1.4831815',
        'min_lat': '45.9667901',
        'min_lon': '-1.4998534',
        'open': False,
        'user': 'randomjunk',
        'uid': 3,
        'tag': {
            'comment': 'correct node bug',
            'created_by': 'Potlatch 1.2a',
        },
    }


def test_ChangesetComment_no_auth(api):
    with pytest.raises(osmapi.UsernamePasswordMissingError) as execinfo:
        api.ChangesetComment(
            123,
            comment="test comment"
        )
    assert str(execinfo.value) == "Username/Password missing"


def test_ChangesetSubscribe(auth_api, add_response):
    add_response(POST, '/changeset/123/subscribe')

    result = auth_api.ChangesetSubscribe(123)

    assert result == {
        'id': 123,
        'closed_at': datetime.datetime(2009, 9, 7, 22, 57, 37),
        'created_at': datetime.datetime(2009, 9, 7, 21, 57, 36),
        'discussion': [],
        'max_lat': '52.4710193',
        'max_lon': '-1.4831815',
        'min_lat': '45.9667901',
        'min_lon': '-1.4998534',
        'open': False,
        'user': 'randomjunk',
        'uid': 3,
        'tag': {
            'comment': 'correct node bug',
            'created_by': 'Potlatch 1.2a',
        },
    }


def test_ChangesetSubscribeWhenAlreadySubscribed(auth_api, add_response):
    add_response(POST, '/changeset/52924/subscribe', status=409)

    with pytest.raises(osmapi.AlreadySubscribedApiError) as execinfo:
        auth_api.ChangesetSubscribe(52924)

    assert execinfo.value.payload == b"You are already subscribed to changeset 52924."
    assert execinfo.value.reason == 'Conflict'
    assert execinfo.value.status == 409


def test_ChangesetSubscribe_no_auth(api):
    with pytest.raises(osmapi.UsernamePasswordMissingError) as execinfo:
        api.ChangesetSubscribe(45627)
    assert str(execinfo.value) == "Username/Password missing"


def test_ChangesetUnsubscribe(auth_api, add_response):
    add_response(POST, '/changeset/123/unsubscribe')

    result = auth_api.ChangesetUnsubscribe(123)

    assert result == {
        'id': 123,
        'closed_at': datetime.datetime(2009, 9, 7, 22, 57, 37),
        'created_at': datetime.datetime(2009, 9, 7, 21, 57, 36),
        'discussion': [],
        'max_lat': '52.4710193',
        'max_lon': '-1.4831815',
        'min_lat': '45.9667901',
        'min_lon': '-1.4998534',
        'open': False,
        'user': 'randomjunk',
        'uid': 3,
        'tag': {
            'comment': 'correct node bug',
            'created_by': 'Potlatch 1.2a',
        },
    }


def test_ChangesetUnsubscribeWhenNotSubscribed(auth_api, add_response):
    add_response(POST, '/changeset/52924/unsubscribe', status=404)

    with pytest.raises(osmapi.NotSubscribedApiError) as execinfo:
        auth_api.ChangesetUnsubscribe(52924)

    assert execinfo.value.payload == b"You are not subscribed to changeset 52924."
    assert execinfo.value.reason == 'Not Found'
    assert execinfo.value.status == 404


def test_ChangesetUnsubscribe_no_auth(api):
    with pytest.raises(osmapi.UsernamePasswordMissingError) as execinfo:
        api.ChangesetUnsubscribe(45627)
    assert str(execinfo.value) == "Username/Password missing"



# -*- coding: utf-8 -*-

"""
The OsmApi module is a wrapper for the OpenStreetMap API.
As such it provides an easy access to the functionality of the API.

You can find this module [on PyPI](https://pypi.python.org/pypi/osmapi)
or [on GitHub](https://github.com/metaodi/osmapi).

Find all information about changes of the different versions of this module
[in the CHANGELOG](https://github.com/metaodi/osmapi/blob/master/CHANGELOG.md).


## Notes:

* **dictionary keys** are _unicode_
* **changeset** is _integer_
* **version** is _integer_
* **tag** is a _dictionary_
* **timestamp** is _unicode_
* **user** is _unicode_
* **uid** is _integer_
* node **lat** and **lon** are _floats_
* way **nd** is list of _integers_
* relation **member** is a _list of dictionaries_ like
`{"role": "", "ref":123, "type": "node"}`

"""

import xml.dom.minidom
import xml.parsers.expat
import urllib.parse
import re
import logging
from contextlib import contextmanager

from osmapi import __version__
from . import dom
from . import errors
from . import http
from . import parser
from . import xmlbuilder


logger = logging.getLogger(__name__)


class OsmApi:
    """
    Main class of osmapi, instanciate this class to use osmapi
    """
    def __init__(
            self,
            username=None,
            password=None,
            passwordfile=None,
            appid="",
            created_by="osmapi/%s" % __version__,
            api="https://www.openstreetmap.org",
            changesetauto=False,
            changesetautotags={},
            changesetautosize=500,
            changesetautomulti=1,
            session=None):
        """
        Initialized the OsmApi object.

        There are two different ways to authenticate a user.
        Either `username` and `password` are supplied directly or the path
        to a `passwordfile` is given, where on the first line username
        and password must be colon-separated (<user>:<pass>).

        To credit the application that supplies changes to OSM, an `appid`
        can be provided.  This is a string identifying the application.
        If this is omitted "osmapi" is used.

        It is possible to configure the URL to connect to using the `api`
        parameter.  By default this is the SSL version of the production API
        of OpenStreetMap, for testing purposes, one might prefer the official
        test instance at "api06.dev.openstreetmap.org" or any other valid
        OSM-API. To use an encrypted connection (HTTPS) simply add 'https://'
        in front of the hostname of the `api` parameter (e.g.
        https://api.openstreetmap.com).

        There are several options to control the changeset behaviour. By
        default, a programmer has to take care to open and close a changeset
        prior to make changes to OSM.
        By setting `changesetauto` to `True`, osmapi automatically opens
        changesets.
        The `changesetautotags` parameter takes a `dict`, where each key/value
        pair is applied as tags to the changeset.
        The option `changesetautosize` defines the size of each
        upload (default: 500) and `changesetautomulti` defines how many
        uploads should be made before closing a changeset and opening a new
        one (default: 1).
        """

        # Get username
        self._username = None
        if username:
            self._username = username
        elif passwordfile:
            with open(passwordfile) as f:
                pass_line = f.readline()
            self._username = pass_line.split(":")[0].strip()

        # Get password
        self._password = None
        if password:
            self._password = password
        elif passwordfile:
            with open(passwordfile) as f:
                for line in f:
                    line = line.strip().split(":", 1)
                    if line[0] == self._username:
                        self._password = line[1]

        # Changest informations
        # auto create and close changesets
        self._changesetauto = changesetauto
        # tags for automatic created changesets
        self._changesetautotags = changesetautotags
        # change count for auto changeset
        self._changesetautosize = changesetautosize
        # change count for auto changeset
        self._changesetautosize = changesetautosize
        # close a changeset every # upload
        self._changesetautomulti = changesetautomulti
        self._changesetautocpt = 0
        # data to upload for auto group
        self._changesetautodata = []

        # Get API
        self._api = api.strip('/')

        # Get created_by
        if not appid:
            self._created_by = created_by
        else:
            self._created_by = "%s (%s)" % (appid, created_by)

        # Initialisation
        self._CurrentChangesetId = 0

        # Http connection
        self.http_session = session
        auth = None
        if self._username and self._password:
            auth = (self._username, self._password)
        self._session = http.OsmApiSession(
            self._api,
            self._created_by,
            auth=auth,
            session=self.http_session
        )

    def __enter__(self):
        self._session = http.OsmApiSession(
            self._api,
            self._created_by,
            session=self.http_session
        )
        return self

    def __exit__(self, *args):
        self.close()

    def close(self):
        try:
            if self._changesetauto:
                self._changesetautoflush(True)
        except errors.ResponseEmptyApiError:
            pass

        if self._session:
            self._session.close()

    ##################################################
    # Capabilities                                   #
    ##################################################

    def Capabilities(self):
        """
        Returns the API capabilities as a dict:

            #!python
            {
                'area': {
                    'maximum': area in square degrees that can be queried,
                },
                'changesets': {
                    'maximum_elements': number of elements per changeset,
                },
                'status': {
                    'api': online|readonly|offline,
                    'database': online|readonly|offline,
                    'gpx': online|readonly|offline,
                },
                'timeout': {
                    'seconds': timeout in seconds for API calls,
                },
                'tracepoints': {
                    'per_page': maximum number of points in a GPX track,
                },
                'version': {
                    'maximum': maximum version of API this server supports,
                    'minimum': minimum version of API this server supports,
                },
                'waynodes': {
                    'maximum': maximum number of nodes that a way may contain,
                },
            }

        The capabilities can be used by a client to
        gain insights of the server in use.
        """
        uri = "/api/capabilities"
        data = self._session._get(uri)

        data = dom.OsmResponseToDom(data, tag="api", single=True)
        result = {}
        for elem in data.childNodes:
            if elem.nodeType != elem.ELEMENT_NODE:
                continue
            result[elem.nodeName] = {}
            for k, v in elem.attributes.items():
                try:
                    result[elem.nodeName][k] = float(v)
                except Exception:
                    result[elem.nodeName][k] = v
        return result

    ##################################################
    # Node                                           #
    ##################################################

    def NodeGet(self, NodeId, NodeVersion=-1):
        """
        Returns node with `NodeId` as a dict:

            #!python
            {
                'id': id of node,
                'lat': latitude of node,
                'lon': longitude of node,
                'tag': {},
                'changeset': id of changeset of last change,
                'version': version number of node,
                'user': username of user that made the last change,
                'uid': id of user that made the last change,
                'timestamp': timestamp of last change,
                'visible': True|False
            }

        If `NodeVersion` is supplied, this specific version is returned,
        otherwise the latest version is returned.

        If the requested element has been deleted,
        `OsmApi.ElementDeletedApiError` is raised.

        If the requested element can not be found,
        `OsmApi.ElementNotFoundApiError` is raised.
        """
        uri = "/api/0.6/node/%s" % (NodeId)
        if NodeVersion != -1:
            uri += "/%s" % (NodeVersion)
        data = self._session._get(uri)
        data = dom.OsmResponseToDom(data, tag="node", single=True)
        return dom.DomParseNode(data)

    def NodeCreate(self, NodeData):
        """
        Creates a node based on the supplied `NodeData` dict:

            #!python
            {
                'lat': latitude of node,
                'lon': longitude of node,
                'tag': {},
            }

        Returns updated `NodeData` (without timestamp):

            #!python
            {
                'id': id of node,
                'lat': latitude of node,
                'lon': longitude of node,
                'tag': dict of tags,
                'changeset': id of changeset of last change,
                'version': version number of node,
                'user': username of last change,
                'uid': id of user of last change,
                'visible': True|False
            }

        If no authentication information are provided,
        `OsmApi.UsernamePasswordMissingError` is raised.

        If there is no open changeset,
        `OsmApi.NoChangesetOpenError` is raised.

        If the supplied information contain an existing node,
        `OsmApi.OsmTypeAlreadyExistsError` is raised.

        If the changeset is already closed,
        `OsmApi.ChangesetClosedApiError` is raised.
        """
        return self._do("create", "node", NodeData)

    def NodeUpdate(self, NodeData):
        """
        Updates node with the supplied `NodeData` dict:

            #!python
            {
                'id': id of node,
                'lat': latitude of node,
                'lon': longitude of node,
                'tag': {},
                'version': version number of node,
            }

        Returns updated `NodeData` (without timestamp):

            #!python
            {
                'id': id of node,
                'lat': latitude of node,
                'lon': longitude of node,
                'tag': dict of tags,
                'changeset': id of changeset of last change,
                'version': version number of node,
                'user': username of last change,
                'uid': id of user of last change,
                'visible': True|False
            }

        If no authentication information are provided,
        `OsmApi.UsernamePasswordMissingError` is raised.

        If there is no open changeset,
        `OsmApi.NoChangesetOpenError` is raised.

        If there is already an open changeset,
        `OsmApi.ChangesetAlreadyOpenError` is raised.

        If the changeset is already closed,
        `OsmApi.ChangesetClosedApiError` is raised.
        """
        return self._do("modify", "node", NodeData)

    def NodeDelete(self, NodeData):
        """
        Delete node with `NodeData`:

            #!python
            {
                'id': id of node,
                'lat': latitude of node,
                'lon': longitude of node,
                'tag': dict of tags,
                'version': version number of node,
            }

        Returns updated `NodeData` (without timestamp):

            #!python
            {
                'id': id of node,
                'lat': latitude of node,
                'lon': longitude of node,
                'tag': dict of tags,
                'changeset': id of changeset of last change,
                'version': version number of node,
                'user': username of last change,
                'uid': id of user of last change,
                'visible': True|False
            }

        If no authentication information are provided,
        `OsmApi.UsernamePasswordMissingError` is raised.

        If there is no open changeset,
        `OsmApi.NoChangesetOpenError` is raised.

        If there is already an open changeset,
        `OsmApi.ChangesetAlreadyOpenError` is raised.

        If the changeset is already closed,
        `OsmApi.ChangesetClosedApiError` is raised.

        If the requested element has already been deleted,
        `OsmApi.ElementDeletedApiError` is raised.

        If the requested element can not be found,
        `OsmApi.ElementNotFoundApiError` is raised.
        """
        return self._do("delete", "node", NodeData)

    def NodeHistory(self, NodeId):
        """
        Returns dict with version as key:

            #!python
            {
                '1': dict of NodeData,
                '2': dict of NodeData,
                ...
            }

        `NodeId` is the unique identifier of a node.
        """
        uri = "/api/0.6/node/%s/history" % NodeId
        data = self._session._get(uri)
        nodes = dom.OsmResponseToDom(data, tag="node")
        result = {}
        for node in nodes:
            data = dom.DomParseNode(node)
            result[data["version"]] = data
        return result

    def NodeWays(self, NodeId):
        """
        Returns a list of dicts of `WayData` containing node `NodeId`:

            #!python
            [
                {
                    'id': id of Way,
                    'nd': [] list of NodeIds in this way
                    'tag': {} dict of tags,
                    'changeset': id of changeset of last change,
                    'version': version number of Way,
                    'user': username of user that made the last change,
                    'uid': id of user that made the last change,
                    'visible': True|False
                },
                {
                    ...
                },
            ]

        The `NodeId` is a unique identifier for a node.
        """
        uri = "/api/0.6/node/%d/ways" % NodeId
        data = self._session._get(uri)
        ways = dom.OsmResponseToDom(data, tag="way", allow_empty=True)
        result = []
        for way in ways:
            data = dom.DomParseWay(way)
            result.append(data)
        return result

    def NodeRelations(self, NodeId):
        """
        Returns a list of dicts of `RelationData` containing node `NodeId`:

            #!python
            [
                {
                    'id': id of Relation,
                    'member': [
                        {
                            'ref': ID of referenced element,
                            'role': optional description of role in relation
                            'type': node|way|relation
                        },
                        {
                            ...
                        }
                    ]
                    'tag': {},
                    'changeset': id of changeset of last change,
                    'version': version number of Way,
                    'user': username of user that made the last change,
                    'uid': id of user that made the last change,
                    'visible': True|False
                },
                {
                    ...
                },
            ]

        The `NodeId` is a unique identifier for a node.
        """
        uri = "/api/0.6/node/%d/relations" % NodeId
        data = self._session._get(uri)
        relations = dom.OsmResponseToDom(data, tag="relation", allow_empty=True)
        result = []
        for relation in relations:
            data = dom.DomParseRelation(relation)
            result.append(data)
        return result

    def NodesGet(self, NodeIdList):
        """
        Returns dict with the id of the Node as a key
        for each node in `NodeIdList`:

            #!python
            {
                '1234': dict of NodeData,
                '5678': dict of NodeData,
                ...
            }

        `NodeIdList` is a list containing unique identifiers
        for multiple nodes.
        """
        node_list = ",".join([str(x) for x in NodeIdList])
        uri = "/api/0.6/nodes?nodes=%s" % node_list
        data = self._session._get(uri)
        nodes = dom.OsmResponseToDom(data, tag="node")
        result = {}
        for node in nodes:
            data = dom.DomParseNode(node)
            result[data["id"]] = data
        return result

    ##################################################
    # Way                                            #
    ##################################################

    def WayGet(self, WayId, WayVersion=-1):
        """
        Returns way with `WayId` as a dict:

            #!python
            {
                'id': id of way,
                'tag': {} tags of this way,
                'nd': [] list of nodes belonging to this way
                'changeset': id of changeset of last change,
                'version': version number of way,
                'user': username of user that made the last change,
                'uid': id of user that made the last change,
                'timestamp': timestamp of last change,
                'visible': True|False
            }

        If `WayVersion` is supplied, this specific version is returned,
        otherwise the latest version is returned.

        If the requested element has been deleted,
        `OsmApi.ElementDeletedApiError` is raised.

        If the requested element can not be found,
        `OsmApi.ElementNotFoundApiError` is raised.
        """
        uri = "/api/0.6/way/%s" % (WayId)
        if WayVersion != -1:
            uri += "/%s" % (WayVersion)
        data = self._session._get(uri)
        way = dom.OsmResponseToDom(data, tag="way", single=True)
        return dom.DomParseWay(way)

    def WayCreate(self, WayData):
        """
        Creates a way based on the supplied `WayData` dict:

            #!python
            {
                'nd': [] list of nodes,
                'tag': {} dict of tags,
            }

        Returns updated `WayData` (without timestamp):

            #!python
            {
                'id': id of node,
                'nd': [] list of nodes,
                'tag': {} dict of tags,
                'changeset': id of changeset of last change,
                'version': version number of way,
                'user': username of last change,
                'uid': id of user of last change,
                'visible': True|False
            }

        If no authentication information are provided,
        `OsmApi.UsernamePasswordMissingError` is raised.

        If the supplied information contain an existing node,
        `OsmApi.OsmTypeAlreadyExistsError` is raised.

        If there is no open changeset,
        `OsmApi.NoChangesetOpenError` is raised.

        If there is already an open changeset,
        `OsmApi.ChangesetAlreadyOpenError` is raised.

        If the changeset is already closed,
        `OsmApi.ChangesetClosedApiError` is raised.
        """
        return self._do("create", "way", WayData)

    def WayUpdate(self, WayData):
        """
        Updates way with the supplied `WayData` dict:

            #!python
            {
                'id': id of way,
                'nd': [] list of nodes,
                'tag': {},
                'version': version number of way,
            }

        Returns updated `WayData` (without timestamp):

            #!python
            {
                'id': id of node,
                'nd': [] list of nodes,
                'tag': {} dict of tags,
                'changeset': id of changeset of last change,
                'version': version number of way,
                'user': username of last change,
                'uid': id of user of last change,
                'visible': True|False
            }

        If no authentication information are provided,
        `OsmApi.UsernamePasswordMissingError` is raised.

        If there is no open changeset,
        `OsmApi.NoChangesetOpenError` is raised.

        If there is already an open changeset,
        `OsmApi.ChangesetAlreadyOpenError` is raised.

        If the changeset is already closed,
        `OsmApi.ChangesetClosedApiError` is raised.
        """
        return self._do("modify", "way", WayData)

    def WayDelete(self, WayData):
        """
        Delete way with `WayData`:

            #!python
            {
                'id': id of way,
                'nd': [] list of nodes,
                'tag': dict of tags,
                'version': version number of way,
            }

        Returns updated `WayData` (without timestamp):

            #!python
            {
                'id': id of node,
                'nd': [] list of nodes,
                'tag': {} dict of tags,
                'changeset': id of changeset of last change,
                'version': version number of way,
                'user': username of last change,
                'uid': id of user of last change,
                'visible': True|False
            }

        If no authentication information are provided,
        `OsmApi.UsernamePasswordMissingError` is raised.

        If there is no open changeset,
        `OsmApi.NoChangesetOpenError` is raised.

        If there is already an open changeset,
        `OsmApi.ChangesetAlreadyOpenError` is raised.

        If the changeset is already closed,
        `OsmApi.ChangesetClosedApiError` is raised.

        If the requested element has already been deleted,
        `OsmApi.ElementDeletedApiError` is raised.

        If the requested element can not be found,
        `OsmApi.ElementNotFoundApiError` is raised.
        """
        return self._do("delete", "way", WayData)

    def WayHistory(self, WayId):
        """
        Returns dict with version as key:

            #!python
            {
                '1': dict of WayData,
                '2': dict of WayData,
                ...
            }

        `WayId` is the unique identifier of a way.
        """
        uri = "/api/0.6/way/%s/history" % (WayId)
        data = self._session._get(uri)
        ways = dom.OsmResponseToDom(data, tag="way")
        result = {}
        for way in ways:
            data = dom.DomParseWay(way)
            result[data["version"]] = data
        return result

    def WayRelations(self, WayId):
        """
        Returns a list of dicts of `RelationData` containing way `WayId`:

            #!python
            [
                {
                    'id': id of Relation,
                    'member': [
                        {
                            'ref': ID of referenced element,
                            'role': optional description of role in relation
                            'type': node|way|relation
                        },
                        {
                            ...
                        }
                    ]
                    'tag': {} dict of tags,
                    'changeset': id of changeset of last change,
                    'version': version number of Way,
                    'user': username of user that made the last change,
                    'uid': id of user that made the last change,
                    'visible': True|False
                },
                {
                    ...
                },
            ]

        The `WayId` is a unique identifier for a way.
        """
        uri = "/api/0.6/way/%d/relations" % WayId
        data = self._session._get(uri)
        relations = dom.OsmResponseToDom(data, tag="relation", allow_empty=True)
        result = []
        for relation in relations:
            data = dom.DomParseRelation(relation)
            result.append(data)
        return result

    def WayFull(self, WayId):
        """
        Returns the full data for way `WayId` as list of dicts:

            #!python
            [
                {
                    'type': node|way|relation,
                    'data': {} data dict for node|way|relation
                },
                { ... }
            ]

        The `WayId` is a unique identifier for a way.

        If the requested element has been deleted,
        `OsmApi.ElementDeletedApiError` is raised.

        If the requested element can not be found,
        `OsmApi.ElementNotFoundApiError` is raised.
        """
        uri = "/api/0.6/way/%s/full" % (WayId)
        data = self._session._get(uri)
        return parser.ParseOsm(data)

    def WaysGet(self, WayIdList):
        """
        Returns dict with the id of the way as a key for
        each way in `WayIdList`:

            #!python
            {
                '1234': dict of WayData,
                '5678': dict of WayData,
                ...
            }

        `WayIdList` is a list containing unique identifiers for multiple ways.
        """
        way_list = ",".join([str(x) for x in WayIdList])
        uri = "/api/0.6/ways?ways=%s" % way_list
        data = self._session._get(uri)
        ways = dom.OsmResponseToDom(data, tag="way")
        result = {}
        for way in ways:
            data = dom.DomParseWay(way)
            result[data["id"]] = data
        return result

    ##################################################
    # Relation                                       #
    ##################################################

    def RelationGet(self, RelationId, RelationVersion=-1):
        """
        Returns relation with `RelationId` as a dict:

            #!python
            {
                'id': id of Relation,
                'member': [
                    {
                        'ref': ID of referenced element,
                        'role': optional description of role in relation
                        'type': node|way|relation
                    },
                    {
                        ...
                    }
                ]
                'tag': {} dict of tags,
                'changeset': id of changeset of last change,
                'version': version number of Relation,
                'user': username of user that made the last change,
                'uid': id of user that made the last change,
                'timestamp': timestamp of last change,
                'visible': True|False
            }

        If `RelationVersion` is supplied, this specific version is returned,
        otherwise the latest version is returned.

        If the requested element has been deleted,
        `OsmApi.ElementDeletedApiError` is raised.

        If the requested element can not be found,
        `OsmApi.ElementNotFoundApiError` is raised.
        """
        uri = "/api/0.6/relation/%s" % (RelationId)
        if RelationVersion != -1:
            uri += "/%s" % (RelationVersion)
        data = self._session._get(uri)
        relation = dom.OsmResponseToDom(data, tag="relation", single=True)
        return dom.DomParseRelation(relation)

    def RelationCreate(self, RelationData):
        """
        Creates a relation based on the supplied `RelationData` dict:

            #!python
            {
                'member': [] list of members,
                'tag': {} dict of tags,
            }

        Returns updated `RelationData` (without timestamp):

            #!python
            {
                'id': id of Relation,
                'member': [
                    {
                        'ref': ID of referenced element,
                        'role': optional description of role in relation
                        'type': node|way|relation
                    },
                    {
                        ...
                    }
                ]
                'tag': {} dict of tags,
                'changeset': id of changeset of last change,
                'version': version number of Relation,
                'user': username of user that made the last change,
                'uid': id of user that made the last change,
                'visible': True|False
            }

        If no authentication information are provided,
        `OsmApi.UsernamePasswordMissingError` is raised.

        If the supplied information contain an existing node,
        `OsmApi.OsmTypeAlreadyExistsError` is raised.

        If there is no open changeset,
        `OsmApi.NoChangesetOpenError` is raised.

        If there is already an open changeset,
        `OsmApi.ChangesetAlreadyOpenError` is raised.

        If the changeset is already closed,
        `OsmApi.ChangesetClosedApiError` is raised.
        """
        return self._do("create", "relation", RelationData)

    def RelationUpdate(self, RelationData):
        """
        Updates relation with the supplied `RelationData` dict:

            #!python
            {
                'id': id of relation,
                'member': [] list of member dicts,
                'tag': {},
                'version': version number of relation,
            }

        Returns updated `RelationData` (without timestamp):

            #!python
            {
                'id': id of Relation,
                'member': [
                    {
                        'ref': ID of referenced element,
                        'role': optional description of role in relation
                        'type': node|way|relation
                    },
                    {
                        ...
                    }
                ]
                'tag': {} dict of tags
                'changeset': id of changeset of last change,
                'version': version number of Relation,
                'user': username of user that made the last change,
                'uid': id of user that made the last change,
                'visible': True|False
            }

        If no authentication information are provided,
        `OsmApi.UsernamePasswordMissingError` is raised.

        If there is no open changeset,
        `OsmApi.NoChangesetOpenError` is raised.

        If there is already an open changeset,
        `OsmApi.ChangesetAlreadyOpenError` is raised.

        If the changeset is already closed,
        `OsmApi.ChangesetClosedApiError` is raised.
        """
        return self._do("modify", "relation", RelationData)

    def RelationDelete(self, RelationData):
        """
        Delete relation with `RelationData` dict:

            #!python
            {
                'id': id of relation,
                'member': [] list of member dicts,
                'tag': {},
                'version': version number of relation,
            }

        Returns updated `RelationData` (without timestamp):

            #!python
            {
                'id': id of Relation,
                'member': [
                    {
                        'ref': ID of referenced element,
                        'role': optional description of role in relation
                        'type': node|way|relation
                    },
                    {
                        ...
                    }
                ]
                'tag': {} dict of tags,
                'changeset': id of changeset of last change,
                'version': version number of Relation,
                'user': username of user that made the last change,
                'uid': id of user that made the last change,
                'visible': True|False
            }

        If no authentication information are provided,
        `OsmApi.UsernamePasswordMissingError` is raised.

        If there is no open changeset,
        `OsmApi.NoChangesetOpenError` is raised.

        If there is already an open changeset,
        `OsmApi.ChangesetAlreadyOpenError` is raised.

        If the changeset is already closed,
        `OsmApi.ChangesetClosedApiError` is raised.

        If the requested element has already been deleted,
        `OsmApi.ElementDeletedApiError` is raised.

        If the requested element can not be found,
        `OsmApi.ElementNotFoundApiError` is raised.
        """
        return self._do("delete", "relation", RelationData)

    def RelationHistory(self, RelationId):
        """
        Returns dict with version as key:

            #!python
            {
                '1': dict of RelationData,
                '2': dict of RelationData,
                ...
            }

        `RelationId` is the unique identifier of a relation.
        """
        uri = "/api/0.6/relation/%s/history" % (RelationId)
        data = self._session._get(uri)
        relations = dom.OsmResponseToDom(data, tag="relation")
        result = {}
        for relation in relations:
            data = dom.DomParseRelation(relation)
            result[data["version"]] = data
        return result

    def RelationRelations(self, RelationId):
        """
        Returns a list of dicts of `RelationData`
        containing relation `RelationId`:

            #!python
            [
                {
                    'id': id of Relation,
                    'member': [
                        {
                            'ref': ID of referenced element,
                            'role': optional description of role in relation
                            'type': node|way|relation
                        },
                        {
                            ...
                        }
                    ]
                    'tag': {} dict of tags,
                    'changeset': id of changeset of last change,
                    'version': version number of Way,
                    'user': username of user that made the last change,
                    'uid': id of user that made the last change,
                    'visible': True|False
                },
                {
                    ...
                },
            ]

        The `RelationId` is a unique identifier for a relation.
        """
        uri = "/api/0.6/relation/%d/relations" % RelationId
        data = self._session._get(uri)
        relations = dom.OsmResponseToDom(data, tag="relation", allow_empty=True)
        result = []
        for relation in relations:
            data = dom.DomParseRelation(relation)
            result.append(data)
        return result

    def RelationFullRecur(self, RelationId):
        """
        Returns the full data (all levels) for relation
        `RelationId` as list of dicts:

            #!python
            [
                {
                    'type': node|way|relation,
                    'data': {} data dict for node|way|relation
                },
                { ... }
            ]

        The `RelationId` is a unique identifier for a way.

        This function is useful for relations containing other relations.

        If you don't need all levels, use `OsmApi.RelationFull`
        instead, which return only 2 levels.

        If any relation (on any level) has been deleted,
        `OsmApi.ElementDeletedApiError` is raised.

        If the requested element can not be found,
        `OsmApi.ElementNotFoundApiError` is raised.
        """
        data = []
        todo = [RelationId]
        done = []
        while todo:
            rid = todo.pop(0)
            done.append(rid)
            temp = self.RelationFull(rid)
            for item in temp:
                if item["type"] != "relation":
                    continue
                if item["data"]["id"] in done:
                    continue
                todo.append(item["data"]["id"])
            data += temp
        return data

    def RelationFull(self, RelationId):
        """
        Returns the full data (two levels) for relation
        `RelationId` as list of dicts:

            #!python
            [
                {
                    'type': node|way|relation,
                    'data': {} data dict for node|way|relation
                },
                { ... }
            ]

        The `RelationId` is a unique identifier for a way.

        If you need all levels, use `OsmApi.RelationFullRecur`.

        If the requested element has been deleted,
        `OsmApi.ElementDeletedApiError` is raised.

        If the requested element can not be found,
        `OsmApi.ElementNotFoundApiError` is raised.
        """
        uri = "/api/0.6/relation/%s/full" % (RelationId)
        data = self._session._get(uri)
        return parser.ParseOsm(data)

    def RelationsGet(self, RelationIdList):
        """
        Returns dict with the id of the relation as a key
        for each relation in `RelationIdList`:

            #!python
            {
                '1234': dict of RelationData,
                '5678': dict of RelationData,
                ...
            }

        `RelationIdList` is a list containing unique identifiers
        for multiple relations.
        """
        relation_list = ",".join([str(x) for x in RelationIdList])
        uri = "/api/0.6/relations?relations=%s" % relation_list
        data = self._session._get(uri)
        relations = dom.OsmResponseToDom(data, tag="relation")
        result = {}
        for relation in relations:
            data = dom.DomParseRelation(relation)
            result[data["id"]] = data
        return result

    ##################################################
    # Changeset                                      #
    ##################################################

    @contextmanager
    def Changeset(self, ChangesetTags={}):
        """
        Context manager for a Changeset.

        It opens a Changeset, uploads the changes and closes the changeset
        when used with the `with` statement:

            #!python
            import osmapi

            with osmapi.Changeset({"comment": "Import script XYZ"}) as changeset_id:
                print(f"Part of changeset {changeset_id}")
                api.NodeCreate({u"lon":1, u"lat":1, u"tag": {}})

        If `ChangesetTags` are given, this tags are applied (key/value).

        Returns `ChangesetId`

        If no authentication information are provided,
        `OsmApi.UsernamePasswordMissingError` is raised.

        If there is already an open changeset,
        `OsmApi.ChangesetAlreadyOpenError` is raised.
        """
        # Create a new changeset
        changeset_id = self.ChangesetCreate(ChangesetTags)
        yield changeset_id

        # upload data to changeset
        autosize = self._changesetautosize
        for i in range(0, len(self._changesetautodata), autosize):
            chunk = self._changesetautodata[i:i+autosize]
            self.ChangesetUpload(chunk)
        self._changesetautodata = []
        self.ChangesetClose()

    def ChangesetGet(self, ChangesetId, include_discussion=False):
        """
        Returns changeset with `ChangesetId` as a dict:

            #!python
            {
                'id': id of Changeset,
                'open': True|False, wheter or not this changeset is open
                'tag': {} dict of tags,
                'created_at': timestamp of creation of this changeset
                'closed_at': timestamp when changeset was closed
                'comments_count': amount of comments
                'discussion': [] list of comment dict (-> `include_discussion`)
                'max_lon': maximum longitude of changes in this changeset
                'max_lat': maximum latitude of changes in this changeset
                'min_lon': minimum longitude of changes in this changeset
                'min_lat': minimum longitude of changes in this changeset
                'user': username of user that created this changeset,
                'uid': id of user that created this changeset,
            }

        `ChangesetId` is the unique identifier of a changeset.

        If `include_discussion` is set to `True` the changeset discussion
        will be available in the result.
        """
        path = "/api/0.6/changeset/%s" % (ChangesetId)
        if (include_discussion):
            path += "?include_discussion=true"
        data = self._session._get(path)
        changeset = dom.OsmResponseToDom(data, tag="changeset", single=True)
        return dom.DomParseChangeset(changeset)

    def ChangesetUpdate(self, ChangesetTags={}):
        """
        Updates current changeset with `ChangesetTags`.

        If no authentication information are provided,
        `OsmApi.UsernamePasswordMissingError` is raised.

        If there is no open changeset,
        `OsmApi.NoChangesetOpenError` is raised.

        If the changeset is already closed,
        `OsmApi.ChangesetClosedApiError` is raised.
        """
        if not self._CurrentChangesetId:
            raise errors.NoChangesetOpenError("No changeset currently opened")
        if "created_by" not in ChangesetTags:
            ChangesetTags["created_by"] = self._created_by
        try:
            self._session._put(
                "/api/0.6/changeset/%s" % (self._CurrentChangesetId),
                xmlbuilder._XmlBuild("changeset", {"tag": ChangesetTags}, data=self),
                return_value=False
            )
        except errors.ApiError as e:
            if e.status == 409:
                raise errors.ChangesetClosedApiError(e.status, e.reason, e.payload)
            else:
                raise
        return self._CurrentChangesetId

    def ChangesetCreate(self, ChangesetTags={}):
        """
        Opens a changeset.

        If `ChangesetTags` are given, this tags are applied (key/value).

        Returns `ChangesetId`

        If no authentication information are provided,
        `OsmApi.UsernamePasswordMissingError` is raised.

        If there is already an open changeset,
        `OsmApi.ChangesetAlreadyOpenError` is raised.
        """
        if self._CurrentChangesetId:
            raise errors.ChangesetAlreadyOpenError("Changeset already opened")
        if "created_by" not in ChangesetTags:
            ChangesetTags["created_by"] = self._created_by

        # check if someone tries to create a test changeset to PROD
        if (
            self._api == "https://www.openstreetmap.org"
            and ChangesetTags.get('comment') == "My first test"
        ):
            raise errors.OsmApiError("DO NOT CREATE test changesets on the production server")

        result = self._session._put(
            "/api/0.6/changeset/create",
            xmlbuilder._XmlBuild("changeset", {"tag": ChangesetTags}, data=self)
        )
        self._CurrentChangesetId = int(result)
        return self._CurrentChangesetId

    def ChangesetClose(self):
        """
        Closes current changeset.

        Returns `ChangesetId`.

        If no authentication information are provided,
        `OsmApi.UsernamePasswordMissingError` is raised.

        If there is no open changeset,
        `OsmApi.NoChangesetOpenError` is raised.

        If the changeset is already closed,
        `OsmApi.ChangesetClosedApiError` is raised.
        """
        if not self._CurrentChangesetId:
            raise errors.NoChangesetOpenError("No changeset currently opened")
        try:
            self._session._put(
                "/api/0.6/changeset/%s/close" % (self._CurrentChangesetId),
                "",
                return_value=False
            )
            CurrentChangesetId = self._CurrentChangesetId
            self._CurrentChangesetId = 0
        except errors.ApiError as e:
            if e.status == 409:
                raise errors.ChangesetClosedApiError(e.status, e.reason, e.payload)
            else:
                raise
        return CurrentChangesetId

    def ChangesetUpload(self, ChangesData):
        """
        Upload data with the `ChangesData` list of dicts:

            #!python
            {
                type: node|way|relation,
                action: create|delete|modify,
                data: {}
            }

        Returns list with updated ids.

        If no authentication information are provided,
        `OsmApi.UsernamePasswordMissingError` is raised.

        If the changeset is already closed,
        `OsmApi.ChangesetClosedApiError` is raised.
        """
        data = ""
        data += "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
        data += "<osmChange version=\"0.6\" generator=\""
        data += self._created_by + "\">\n"
        for change in ChangesData:
            data += "<" + change["action"] + ">\n"
            change["data"]["changeset"] = self._CurrentChangesetId
            data += xmlbuilder._XmlBuild(
                change["type"],
                change["data"],
                False,
                data=self
            ).decode("utf-8")
            data += "</" + change["action"] + ">\n"
        data += "</osmChange>"
        try:
            data = self._session._post(
                "/api/0.6/changeset/%s/upload" % (self._CurrentChangesetId),
                data.encode("utf-8"),
                forceAuth=True
            )
        except errors.ApiError as e:
            if e.status == 409 and re.search(r"The changeset .* was closed at .*", e.payload):
                raise errors.ChangesetClosedApiError(e.status, e.reason, e.payload)
            else:
                raise
        try:
            data = xml.dom.minidom.parseString(data)
            data = data.getElementsByTagName("diffResult")[0]
            data = [x for x in data.childNodes if x.nodeType == x.ELEMENT_NODE]
        except (xml.parsers.expat.ExpatError, IndexError) as e:
            raise errors.XmlResponseInvalidError(
                "The XML response from the OSM API is invalid: %r" % e
            )

        for i in range(len(ChangesData)):
            if ChangesData[i]["action"] == "delete":
                ChangesData[i]["data"].pop("version")
            else:
                new_id = int(data[i].getAttribute("new_id"))
                ChangesData[i]["data"]["id"] = new_id
                new_version = int(data[i].getAttribute("new_version"))
                ChangesData[i]["data"]["version"] = new_version
        return ChangesData

    def ChangesetDownload(self, ChangesetId):
        """
        Download data from changeset `ChangesetId`.

        Returns list of dict:

            #!python
            {
                'type': node|way|relation,
                'action': create|delete|modify,
                'data': {}
            }
        """
        uri = "/api/0.6/changeset/%s/download" % (ChangesetId)
        data = self._session._get(uri)
        return parser.ParseOsc(data)

    def ChangesetsGet(  # noqa
            self,
            min_lon=None,
            min_lat=None,
            max_lon=None,
            max_lat=None,
            userid=None,
            username=None,
            closed_after=None,
            created_before=None,
            only_open=False,
            only_closed=False):
        """
        Returns a dict with the id of the changeset as key
        matching all criteria:

            #!python
            {
                '1234': dict of ChangesetData,
                '5678': dict of ChangesetData,
                ...
            }

        All parameters are optional.
        """

        uri = "/api/0.6/changesets"
        params = {}
        if min_lon or min_lat or max_lon or max_lat:
            params["bbox"] = ",".join(
                [
                    str(min_lon),
                    str(min_lat),
                    str(max_lon),
                    str(max_lat)
                ]
            )
        if userid:
            params["user"] = userid
        if username:
            params["display_name"] = username
        if closed_after and not created_before:
            params["time"] = closed_after
        if created_before:
            if not closed_after:
                closed_after = "1970-01-01T00:00:00Z"
            params["time"] = "%s,%s" % (closed_after, created_before)
        if only_open:
            params["open"] = 1
        if only_closed:
            params["closed"] = 1

        if params:
            uri += "?" + urllib.parse.urlencode(params)

        data = self._session._get(uri)
        changesets = dom.OsmResponseToDom(data, tag="changeset")
        result = {}
        for curChangeset in changesets:
            tmpCS = dom.DomParseChangeset(curChangeset)
            result[tmpCS["id"]] = tmpCS
        return result

    def ChangesetComment(self, ChangesetId, comment):
        """
        Adds a comment to the changeset `ChangesetId`

        `comment` should be a string.

        Returns the updated `ChangesetData` dict:

            #!python
            {
                'id': id of Changeset,
                'open': True|False, wheter or not this changeset is open
                'tag': {} dict of tags,
                'created_at': timestamp of creation of this changeset
                'closed_at': timestamp when changeset was closed
                'comments_count': amount of comments
                'max_lon': maximum longitude of changes in this changeset
                'max_lat': maximum latitude of changes in this changeset
                'min_lon': minimum longitude of changes in this changeset
                'min_lat': minimum longitude of changes in this changeset
                'user': username of user that created this changeset,
                'uid': id of user that created this changeset,
            }


        If no authentication information are provided,
        `OsmApi.UsernamePasswordMissingError` is raised.

        If the changeset is already closed,
        `OsmApi.ChangesetClosedApiError` is raised.
        """
        params = urllib.parse.urlencode({'text': comment})
        try:
            data = self._session._post(
                "/api/0.6/changeset/%s/comment" % (ChangesetId),
                params,
                forceAuth=True
            )
        except errors.ApiError as e:
            if e.status == 409:
                raise errors.ChangesetClosedApiError(e.status, e.reason, e.payload)
            else:
                raise
        changeset = dom.OsmResponseToDom(data, tag="changeset", single=True)
        return dom.DomParseChangeset(changeset)

    def ChangesetSubscribe(self, ChangesetId):
        """
        Subcribe to the changeset discussion of changeset `ChangesetId`.

        The user will be informed about new comments (i.e. receive an email).

        Returns the updated `ChangesetData` dict:

            #!python
            {
                'id': id of Changeset,
                'open': True|False, wheter or not this changeset is open
                'tag': {} dict of tags,
                'created_at': timestamp of creation of this changeset
                'closed_at': timestamp when changeset was closed
                'comments_count': amount of comments
                'max_lon': maximum longitude of changes in this changeset
                'max_lat': maximum latitude of changes in this changeset
                'min_lon': minimum longitude of changes in this changeset
                'min_lat': minimum longitude of changes in this changeset
                'user': username of user that created this changeset,
                'uid': id of user that created this changeset,
            }

        If no authentication information are provided,
        `OsmApi.UsernamePasswordMissingError` is raised.
        """
        try:
            data = self._session._post(
                "/api/0.6/changeset/%s/subscribe" % (ChangesetId),
                None,
                forceAuth=True
            )
        except errors.ApiError as e:
            if e.status == 409:
                raise errors.AlreadySubscribedApiError(e.status, e.reason, e.payload)
            else:
                raise
        changeset = dom.OsmResponseToDom(data, tag="changeset", single=True)
        return dom.DomParseChangeset(changeset)

    def ChangesetUnsubscribe(self, ChangesetId):
        """
        Subcribe to the changeset discussion of changeset `ChangesetId`.

        The user will be informed about new comments (i.e. receive an email).

        Returns the updated `ChangesetData` dict:

            #!python
            {
                'id': id of Changeset,
                'open': True|False, wheter or not this changeset is open
                'tag': {} dict of tags,
                'created_at': timestamp of creation of this changeset
                'closed_at': timestamp when changeset was closed
                'comments_count': amount of comments
                'max_lon': maximum longitude of changes in this changeset
                'max_lat': maximum latitude of changes in this changeset
                'min_lon': minimum longitude of changes in this changeset
                'min_lat': minimum longitude of changes in this changeset
                'user': username of user that created this changeset,
                'uid': id of user that created this changeset,
            }

        If no authentication information are provided,
        `OsmApi.UsernamePasswordMissingError` is raised.
        """
        try:
            data = self._session._post(
                "/api/0.6/changeset/%s/unsubscribe" % (ChangesetId),
                None,
                forceAuth=True
            )
        except errors.ElementNotFoundApiError as e:
            raise errors.NotSubscribedApiError(e.status, e.reason, e.payload)

        changeset = dom.OsmResponseToDom(data, tag="changeset", single=True)
        return dom.DomParseChangeset(changeset)

    ##################################################
    # Notes                                          #
    ##################################################

    def NotesGet(
            self,
            min_lon,
            min_lat,
            max_lon,
            max_lat,
            limit=100,
            closed=7):
        """
        Returns a list of dicts of notes in the specified bounding box:

            #!python
            [
                {
                    'id': integer,
                    'action': opened|commented|closed,
                    'status': open|closed
                    'date_created': creation date
                    'date_closed': closing data|None
                    'uid': User ID|None
                    'user': User name|None
                    'comments': {}
                },
                { ... }
            ]

        The limit parameter defines how many results should be returned.

        closed specifies the number of days a bug needs to be closed
        to no longer be returned.
        The value 0 means only open bugs are returned,
        -1 means all bugs are returned.

        All parameters are optional.
        """
        uri = (
            "/api/0.6/notes?bbox=%f,%f,%f,%f&limit=%d&closed=%d"
            % (min_lon, min_lat, max_lon, max_lat, limit, closed)
        )
        data = self._session._get(uri)
        return parser.ParseNotes(data)

    def NoteGet(self, id):
        """
        Returns a note as dict:

            #!python
            {
                'id': integer,
                'action': opened|commented|closed,
                'status': open|closed
                'date_created': creation date
                'date_closed': closing data|None
                'uid': User ID|None
                'user': User name|None
                'comments': {}
            }

        `id` is the unique identifier of the note.
        """
        uri = "/api/0.6/notes/%s" % (id)
        data = self._session._get(uri)
        noteElement = dom.OsmResponseToDom(data, tag="note", single=True)
        return dom.DomParseNote(noteElement)

    def NoteCreate(self, NoteData):
        """
        Creates a note based on the supplied `NoteData` dict:

            #!python
            {
                'lat': latitude of note,
                'lon': longitude of note,
                'text': text of the note,
            }

        Returns updated `NoteData`:

            #!python
            {
                'id': id of note,
                'lat': latitude of note,
                'lon': longitude of note,
                'date_created': date when the note was created
                'date_closed': date when the note was closed (or None if the note is open),
                'status': status of the note (open or closed),
                'comments': [
                    {
                        'date': date of the comment,
                        'action': status of comment (opened, commented, closed),
                        'text': text of the note,
                        'html': html version of the text of the note,
                        'uid': user id of the user creating this note or None
                        'user': username of the user creating this note or None
                    }
                ]
            }

        """
        uri = "/api/0.6/notes"
        uri += "?" + urllib.parse.urlencode(NoteData)
        return self._NoteAction(uri)

    def NoteComment(self, NoteId, comment):
        """
        Adds a new comment to a note.

        Returns the updated note.
        """
        path = "/api/0.6/notes/%s/comment" % NoteId
        return self._NoteAction(path, comment)

    def NoteClose(self, NoteId, comment):
        """
        Closes a note.

        Returns the updated note.

        If no authentication information are provided,
        `OsmApi.UsernamePasswordMissingError` is raised.
        """
        path = "/api/0.6/notes/%s/close" % NoteId
        return self._NoteAction(path, comment, optionalAuth=False)

    def NoteReopen(self, NoteId, comment):
        """
        Reopens a note.

        Returns the updated note.

        If no authentication information are provided,
        `OsmApi.UsernamePasswordMissingError` is raised.

        If the requested element has been deleted,
        `OsmApi.ElementDeletedApiError` is raised.

        If the requested element can not be found,
        `OsmApi.ElementNotFoundApiError` is raised.
        """
        path = "/api/0.6/notes/%s/reopen" % NoteId
        return self._NoteAction(path, comment, optionalAuth=False)

    def NotesSearch(self, query, limit=100, closed=7):
        """
        Returns a list of dicts of notes that match the given search query.

        The limit parameter defines how many results should be returned.

        closed specifies the number of days a bug needs to be closed
        to no longer be returned.
        The value 0 means only open bugs are returned,
        -1 means all bugs are returned.
        """
        uri = "/api/0.6/notes/search"
        params = {}
        params['q'] = query
        params['limit'] = limit
        params['closed'] = closed
        uri += "?" + urllib.parse.urlencode(params)
        data = self._session._get(uri)

        return parser.ParseNotes(data)

    def _NoteAction(self, path, comment=None, optionalAuth=True):
        """
        Performs an action on a Note with a comment

        Return the updated note
        """
        uri = path
        if comment is not None:
            params = {}
            params['text'] = comment
            uri += "?" + urllib.parse.urlencode(params)
        try:
            result = self._session._post(uri, None, optionalAuth=optionalAuth)
        except errors.ApiError as e:
            if e.status == 409:
                raise errors.NoteAlreadyClosedApiError(e.status, e.reason, e.payload)
            else:
                raise

        # parse the result
        noteElement = dom.OsmResponseToDom(result, tag="note", single=True)
        return dom.DomParseNote(noteElement)

    ##################################################
    # Other                                          #
    ##################################################

    def Map(self, min_lon, min_lat, max_lon, max_lat):
        """
        Download data in bounding box.

        Returns list of dict:

            #!python
            {
                type: node|way|relation,
                data: {}
            }
        """
        uri = (
            "/api/0.6/map?bbox=%f,%f,%f,%f"
            % (min_lon, min_lat, max_lon, max_lat)
        )
        data = self._session._get(uri)
        return parser.ParseOsm(data)

    def flush(self):
        """
        Force the changes to be uploaded to OSM and the changeset to be closed

        If no authentication information are provided,
        `OsmApi.UsernamePasswordMissingError` is raised.

        If there is no open changeset,
        `OsmApi.NoChangesetOpenError` is raised.

        If there is already an open changeset,
        `OsmApi.ChangesetAlreadyOpenError` is raised.
        """
        return self._changesetautoflush(True)

    ##################################################
    # Internal method                                #
    ##################################################

    def _do(self, action, OsmType, OsmData):
        if self._changesetauto:
            self._changesetautodata.append({
                "action": action,
                "type": OsmType,
                "data": OsmData
            })
            self._changesetautoflush()
            return None
        else:
            return self._do_manu(action, OsmType, OsmData)

    def _do_manu(self, action, OsmType, OsmData):  # noqa
        if not self._CurrentChangesetId:
            raise errors.NoChangesetOpenError(
                "You need to open a changeset before uploading data"
            )
        if "timestamp" in OsmData:
            OsmData.pop("timestamp")
        OsmData["changeset"] = self._CurrentChangesetId
        if action == "create":
            if OsmData.get("id", -1) > 0:
                raise errors.OsmTypeAlreadyExistsError(
                    "This %s already exists" % OsmType
                )
            try:
                result = self._session._put(
                    "/api/0.6/%s/create" % OsmType,
                    xmlbuilder._XmlBuild(OsmType, OsmData, data=self)
                )
            except errors.ApiError as e:
                if e.status == 409 and re.search(r"The changeset .* was closed at .*", e.payload):
                    raise errors.ChangesetClosedApiError(e.status, e.reason, e.payload)
                elif e.status == 409:
                    raise errors.VersionMismatchApiError(e.status, e.reason, e.payload)
                elif e.status == 412:
                    raise errors.PreconditionFailedApiError(e.status, e.reason, e.payload)
                else:
                    raise
            OsmData["id"] = int(result.strip())
            OsmData["version"] = 1
            return OsmData
        elif action == "modify":
            try:
                result = self._session._put(
                    "/api/0.6/%s/%s" % (OsmType, OsmData["id"]),
                    xmlbuilder._XmlBuild(OsmType, OsmData, data=self)
                )
            except errors.ApiError as e:
                logger.error(e.reason)
                if e.status == 409 and re.search(r"The changeset .* was closed at .*", e.payload):
                    raise errors.ChangesetClosedApiError(e.status, e.reason, e.payload)
                elif e.status == 409:
                    raise errors.VersionMismatchApiError(e.status, e.reason, e.payload)
                elif e.status == 412:
                    raise errors.PreconditionFailedApiError(e.status, e.reason, e.payload)
                else:
                    raise
            OsmData["version"] = int(result.strip())
            return OsmData
        elif action == "delete":
            try:
                result = self._session._delete(
                    "/api/0.6/%s/%s" % (OsmType, OsmData["id"]),
                    xmlbuilder._XmlBuild(OsmType, OsmData, data=self)
                )
            except errors.ApiError as e:
                if e.status == 409 and re.search(r"The changeset .* was closed at .*", e.payload):
                    raise errors.ChangesetClosedApiError(e.status, e.reason, e.payload)
                elif e.status == 409:
                    raise errors.VersionMismatchApiError(e.status, e.reason, e.payload)
                elif e.status == 412:
                    raise errors.PreconditionFailedApiError(e.status, e.reason, e.payload)
                else:
                    raise
            OsmData["version"] = int(result.strip())
            OsmData["visible"] = False
            return OsmData

    def _changesetautoflush(self, force=False):
        autosize = self._changesetautosize
        while ((len(self._changesetautodata) >= autosize) or
                (force and self._changesetautodata)):
            if self._changesetautocpt == 0:
                self.ChangesetCreate(self._changesetautotags)
            self.ChangesetUpload(
                self._changesetautodata[:autosize]
            )
            self._changesetautodata = self._changesetautodata[autosize:]
            self._changesetautocpt += 1
            if self._changesetautocpt == self._changesetautomulti:
                self.ChangesetClose()
                self._changesetautocpt = 0
        if self._changesetautocpt and force:
            self.ChangesetClose()
            self._changesetautocpt = 0
        return None



from __future__ import (unicode_literals, absolute_import)
from . import osmapi_test
import osmapi
import mock
import os

__location__ = os.path.realpath(
    os.path.join(
        os.getcwd(),
        os.path.dirname(__file__)
    )
)


class TestOsmApiHelper(osmapi_test.TestOsmApi):
    def setUp(self):
        super(TestOsmApiHelper, self).setUp()
        self.setupMock()

    def setupMock(self, status=200):
        mock_response = mock.Mock()
        mock_response.status_code = status
        mock_response.reason = "test reason"
        mock_response.content = 'test response'

        self.mock_session = mock.Mock()
        self.mock_session.request = mock.Mock(return_value=mock_response)
        self.mock_session.close = mock.Mock()
        self.mock_session.auth = ('testuser', 'testpassword')

        self.api = osmapi.OsmApi(
            api=self.api_base,
            session=self.mock_session,
            username='testuser',
            password='testpassword'
        )

    def test_passwordfile_only(self):
        path = os.path.join(
            __location__,
            'fixtures',
            'passwordfile.txt'
        )
        my_api = osmapi.OsmApi(passwordfile=path)
        self.assertEqual('testosm', my_api._username)
        self.assertEqual('testpass', my_api._password)

    def test_passwordfile_with_user(self):
        path = os.path.join(
            __location__,
            'fixtures',
            'passwordfile.txt'
        )
        my_api = osmapi.OsmApi(username='testuser', passwordfile=path)
        self.assertEqual('testuser', my_api._username)
        self.assertEqual('testuserpass', my_api._password)

    def test_passwordfile_with_colon(self):
        path = os.path.join(
            __location__,
            'fixtures',
            'passwordfile_colon.txt'
        )
        my_api = osmapi.OsmApi(username='testuser', passwordfile=path)
        self.assertEqual('testuser', my_api._username)
        self.assertEqual('test:userpass', my_api._password)

    def test_close_call(self):
        self.api.close()
        self.assertEqual(self.api._session._session.close.call_count, 1)

    def test_close_context_manager(self):
        with osmapi.OsmApi() as my_api:
            my_api._session.close = mock.Mock()
        self.assertEqual(my_api._session.close.call_count, 1)

    def test_http_request_get(self):
        response = self.api._session._http_request(
            'GET',
            '/api/0.6/test',
            False,
            None
        )
        self.mock_session.request.assert_called_with(
            'GET',
            self.api_base + '/api/0.6/test',
            data=None
        )
        self.assertEqual(response, "test response")
        self.assertEqual(self.mock_session.request.call_count, 1)

    def test_http_request_put(self):
        data = "data"
        response = self.api._session._http_request(
            'PUT',
            '/api/0.6/testput',
            False,
            data
        )
        self.mock_session.request.assert_called_with(
            'PUT',
            self.api_base + '/api/0.6/testput',
            data="data"
        )
        self.assertEqual(response, "test response")

    def test_http_request_delete(self):
        data = "delete data"
        response = self.api._session._http_request(
            'PUT',
            '/api/0.6/testdelete',
            False,
            data
        )
        self.mock_session.request.assert_called_with(
            'PUT',
            self.api_base + '/api/0.6/testdelete',
            data="delete data"
        )
        self.assertEqual(response, "test response")

    def test_http_request_auth(self):
        response = self.api._session._http_request(
            'PUT',
            '/api/0.6/testauth',
            True,
            None
        )
        self.mock_session.request.assert_called_with(
            'PUT',
            self.api_base + '/api/0.6/testauth',
            data=None
        )
        self.assertEqual(self.mock_session.auth, ('testuser', 'testpassword'))
        self.assertEqual(response, "test response")

    def test_http_request_410_response(self):
        self.setupMock(410)
        with self.assertRaises(osmapi.ElementDeletedApiError) as cm:
            self.api._session._http_request(
                'GET',
                '/api/0.6/test410',
                False,
                None
            )
        self.assertEqual(cm.exception.status, 410)
        self.assertEqual(cm.exception.reason, "test reason")
        self.assertEqual(cm.exception.payload, "test response")

    def test_http_request_500_response(self):
        self.setupMock(500)
        with self.assertRaises(osmapi.ApiError) as cm:
            self.api._session._http_request(
                'GET',
                self.api_base + '/api/0.6/test500',
                False,
                None
            )
        self.assertEqual(cm.exception.status, 500)
        self.assertEqual(cm.exception.reason, "test reason")
        self.assertEqual(cm.exception.payload, "test response")



from . import osmapi_test
import osmapi
import mock
import datetime


class TestOsmApiDom(osmapi_test.TestOsmApi):
    def test_DomGetAttributes(self):
        mock_domelement = mock.Mock()
        mock_domelement.attributes = {
            'uid': '12345',
            'open': 'false',
            'visible': 'true',
            'lat': '47.1234',
            'date': '2021-12-10T21:28:03Z',
            'new_attribute': 'Test 123',
        }

        result = osmapi.dom._DomGetAttributes(mock_domelement)

        self.assertIsInstance(result, dict)
        self.assertEqual(result['uid'], 12345)
        self.assertEqual(result['open'], False)
        self.assertEqual(result['visible'], True)
        self.assertEqual(result['lat'], 47.1234)
        self.assertEqual(result['date'], datetime.datetime(2021, 12, 10, 21, 28, 3))
        self.assertEqual(result['new_attribute'], 'Test 123')

    def test_ParseDate(self):
        self.assertEqual(
            osmapi.dom._ParseDate('2021-02-25T09:49:33Z'),
            datetime.datetime(2021, 2, 25, 9, 49, 33)
        )
        self.assertEqual(
            osmapi.dom._ParseDate('2021-02-25 09:49:33 UTC'),
            datetime.datetime(2021, 2, 25, 9, 49, 33)
        )
        with self.assertLogs('osmapi.dom', level='DEBUG') as cm:
            self.assertEqual(
                osmapi.dom._ParseDate('2021-02-25'),
                '2021-02-25'
            )
            self.assertEqual(
                osmapi.dom._ParseDate(''),
                ''
            )
            self.assertIsNone(osmapi.dom._ParseDate(None))

            # test logging output
            self.assertEqual(
                cm.output,
                [
                    'DEBUG:osmapi.dom:2021-02-25 does not match %Y-%m-%d %H:%M:%S UTC',
                    'DEBUG:osmapi.dom:2021-02-25 does not match %Y-%m-%dT%H:%M:%SZ',
                    'DEBUG:osmapi.dom: does not match %Y-%m-%d %H:%M:%S UTC',
                    'DEBUG:osmapi.dom: does not match %Y-%m-%dT%H:%M:%SZ',
                    'DEBUG:osmapi.dom:None does not match %Y-%m-%d %H:%M:%S UTC',
                    'DEBUG:osmapi.dom:None does not match %Y-%m-%dT%H:%M:%SZ',
                ]
            )



from __future__ import (unicode_literals, absolute_import)
from . import osmapi_test


class TestOsmApiNode(osmapi_test.TestOsmApi):
    def test_Capabilities(self):
        self._session_mock()

        result = self.api.Capabilities()
        self.assertEqual(result, {
            'area': {'maximum': 0.25},
            'changesets': {'maximum_elements': 50000.0},
            'status': {
                'api': 'mocked',
                'database': 'online',
                'gpx': 'online'
            },
            'timeout': {'seconds': 300.0},
            'tracepoints': {'per_page': 5000.0},
            'version': {'maximum': 0.6, 'minimum': 0.6},
            'waynodes': {'maximum': 2000.0}
        })



from __future__ import (unicode_literals, absolute_import)
from . import osmapi_test
import osmapi
import mock
import datetime


class TestOsmApiWay(osmapi_test.TestOsmApi):
    def test_WayGet(self):
        self._session_mock()

        result = self.api.WayGet(321)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], 'GET')
        self.assertEqual(args[1], self.api_base + '/api/0.6/way/321')

        self.assertEqual(result, {
            'id': 321,
            'changeset': 298,
            'uid': 12,
            'timestamp': datetime.datetime(2009, 9, 14, 23, 23, 18),
            'visible': True,
            'version': 1,
            'user': 'green525',
            'tag': {
                'admin_level': '9',
                'boundary': 'administrative',
            },
            'nd': [
                11949,
                11950,
                11951,
                11952,
                11953,
                11954,
                11955,
                11956,
                11957,
                11958,
                11959,
                11960,
                11961,
                11962,
                11963,
                11964,
                11949
            ]
        })

    def test_WayGet_with_version(self):
        self._session_mock()

        result = self.api.WayGet(4294967296, 2)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], 'GET')
        self.assertEqual(
            args[1],
            f'{self.api_base}/api/0.6/way/4294967296/2'
        )

        self.assertEqual(result['id'], 4294967296)
        self.assertEqual(result['changeset'], 41303)
        self.assertEqual(result['user'], 'metaodi')

    def test_WayGet_nodata(self):
        self._session_mock()

        with self.assertRaises(osmapi.ResponseEmptyApiError):
            self.api.WayGet(321)

    def test_WayCreate(self):
        self._session_mock(auth=True)

        # setup mock
        self.api.ChangesetCreate = mock.Mock(
            return_value=2222
        )
        self.api._CurrentChangesetId = 2222

        test_way = {
            'nd': [11949, 11950],
            'tag': {
                'highway': 'unclassified',
                'name': 'Osmapi Street'
            }
        }

        cs = self.api.ChangesetCreate({
            'comment': 'This is a test way'
        })
        self.assertEqual(cs, 2222)

        result = self.api.WayCreate(test_way)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], 'PUT')
        self.assertEqual(args[1], self.api_base + '/api/0.6/way/create')

        self.assertEqual(result['id'], 5454)
        self.assertEqual(result['nd'], test_way['nd'])
        self.assertEqual(result['tag'], test_way['tag'])

    def test_WayCreate_existing_node(self):
        # setup mock
        self.api.ChangesetCreate = mock.Mock(
            return_value=1111
        )
        self.api._CurrentChangesetId = 1111

        test_way = {
            'id': 456,
            'nd': [11949, 11950],
            'tag': {
                'highway': 'unclassified',
                'name': 'Osmapi Street'
            }
        }

        with self.assertRaisesRegex(
                osmapi.OsmTypeAlreadyExistsError,
                'This way already exists'):
            self.api.WayCreate(test_way)

    def test_WayUpdate(self):
        self._session_mock(auth=True)

        # setup mock
        self.api.ChangesetCreate = mock.Mock(
            return_value=2222
        )
        self.api._CurrentChangesetId = 2222

        test_way = {
            'id': 876,
            'nd': [11949, 11950],
            'tag': {
                'highway': 'unclassified',
                'name': 'Osmapi Street Update'
            }
        }

        cs = self.api.ChangesetCreate({
            'comment': 'This is a test way'
        })
        self.assertEqual(cs, 2222)

        result = self.api.WayUpdate(test_way)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], 'PUT')
        self.assertEqual(args[1], self.api_base + '/api/0.6/way/876')

        self.assertEqual(result['id'], 876)
        self.assertEqual(result['version'], 7)
        self.assertEqual(result['nd'], test_way['nd'])
        self.assertEqual(result['tag'], test_way['tag'])

    def test_WayUpdatePreconditionFailed(self):
        self._session_mock(auth=True, status=412)

        self.api.ChangesetCreate = mock.Mock(
            return_value=1111
        )
        self.api._CurrentChangesetId = 1111

        test_way = {
            'id': 876,
            'nd': [11949, 11950],
            'tag': {
                'highway': 'unclassified',
                'name': 'Osmapi Street Update'
            }
        }

        self.api.ChangesetCreate({
            'comment': 'This is a test dataset'
        })

        with self.assertRaises(osmapi.PreconditionFailedApiError) as cm:
            self.api.WayUpdate(test_way)

        self.assertEqual(cm.exception.status, 412)
        self.assertEqual(
            cm.exception.payload,
            (
                "Way 876 requires the nodes with id in (11950), "
                "which either do not exist, or are not visible."
            )
        )

    def test_WayDelete(self):
        self._session_mock(auth=True)

        # setup mock
        self.api.ChangesetCreate = mock.Mock(
            return_value=2222
        )
        self.api._CurrentChangesetId = 2222

        test_way = {
            'id': 876
        }

        cs = self.api.ChangesetCreate({
            'comment': 'This is a test way delete'
        })
        self.assertEqual(cs, 2222)

        result = self.api.WayDelete(test_way)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], 'DELETE')
        self.assertEqual(args[1], self.api_base + '/api/0.6/way/876')
        self.assertEqual(result['id'], 876)
        self.assertEqual(result['version'], 8)

    def test_WayHistory(self):
        self._session_mock()

        result = self.api.WayHistory(4294967296)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], 'GET')
        self.assertEqual(
            args[1],
            f'{self.api_base}/api/0.6/way/4294967296/history'
        )

        self.assertEqual(len(result), 2)
        self.assertEqual(result[1]['id'], 4294967296)
        self.assertEqual(result[1]['version'], 1)
        self.assertEqual(
            result[1]['tag'], {
                'highway': 'unclassified',
                'name': 'Stansted Road',
            }
        )

    def test_WayRelations(self):
        self._session_mock()

        result = self.api.WayRelations(4295032193)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], 'GET')
        self.assertEqual(
            args[1],
            f'{self.api_base}/api/0.6/way/4295032193/relations'
        )

        self.assertEqual(len(result), 1)
        self.assertEqual(result[0]['id'], 4294968148)
        self.assertEqual(result[0]['changeset'], 23123)
        self.assertEqual(
            result[0]['member'][4],
            {
                'role': '',
                'ref': 4295032193,
                'type': 'way',
            }
        )
        self.assertEqual(
            result[0]['tag'],
            {
                'type': 'fancy',
            }
        )

    def test_WayRelationsUnusedElement(self):
        self._session_mock()

        result = self.api.WayRelations(4295032193)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], 'GET')
        self.assertEqual(
            args[1],
            self.api_base + '/api/0.6/way/4295032193/relations'
        )

        self.assertEqual(len(result), 0)
        self.assertIsInstance(result, list)

    def test_WayFull(self):
        self._session_mock()

        result = self.api.WayFull(321)

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], 'GET')
        self.assertEqual(args[1], self.api_base + '/api/0.6/way/321/full')

        self.assertEqual(len(result), 17)
        self.assertEqual(result[0]['data']['id'], 11949)
        self.assertEqual(result[0]['data']['changeset'], 298)
        self.assertEqual(result[0]['type'], 'node')
        self.assertEqual(result[16]['data']['id'], 321)
        self.assertEqual(result[16]['data']['changeset'], 298)
        self.assertEqual(result[16]['type'], 'way')

    def test_WayFull_invalid_response(self):
        self._session_mock()

        with self.assertRaises(osmapi.XmlResponseInvalidError):
            self.api.WayFull(321)

    def test_WaysGet(self):
        self._session_mock()

        result = self.api.WaysGet([456, 678])

        args, kwargs = self.session_mock.request.call_args
        self.assertEqual(args[0], 'GET')
        self.assertEqual(
            args[1],
            f'{self.api_base}/api/0.6/ways?ways=456,678'
        )

        self.assertEqual(len(result), 2)
        self.assertIs(type(result[456]), dict)
        self.assertIs(type(result[678]), dict)
        with self.assertRaises(KeyError):
            self.assertIs(type(result[123]), dict)



# -*- coding: utf-8 -*-

from codecs import open
from setuptools import setup, find_packages
import re

with open('osmapi/__init__.py', 'r') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError('Cannot find version information')

with open('README.md', 'r', encoding="utf-8") as f:
    long_description = f.read()

setup(
    name='osmapi',
    packages=find_packages(),
    version=version,
    install_requires=['requests'],
    python_requires='>=3.7',
    description='Python wrapper for the OSM API',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Etienne Chové',
    author_email='chove@crans.org',
    maintainer='Stefan Oderbolz',
    maintainer_email='odi@metaodi.ch',
    url='https://github.com/metaodi/osmapi',
    download_url='https://github.com/metaodi/osmapi/archive/v%s.zip' % version,
    keywords=['openstreetmap', 'osm', 'api'],
    license='GPLv3',
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: GIS',
        'Topic :: Software Development :: Libraries',
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
)
